#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import seaborn as sns

from kxy.api.core import auto_predictability

from kxy.classification import classification_achievable_performance_analysis, \
	classification_variable_selection_analysis

from kxy.regression import regression_achievable_performance_analysis, \
	regression_variable_selection_analysis

from .base_accessor import BaseAccessor


@pd.api.extensions.register_dataframe_accessor("kxy_pre_learning")
class PreLearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics **pre-learning** in supervised learning problems.

	This class defines the :code:`kxy_pre_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_pre_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	def achievable_performance_analysis(self, label_column, input_columns=(), space='dual', categorical_encoding='two-split', problem=None):
		"""
		Runs the achievable performance analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.



		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

			* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a model using provided inputs to predict the label.



		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.

		.. seealso::

			* :ref:`kxy.regression.regression_achievable_performance_analysis <regression-achievable-performance-analysis>`
			* :ref:`kxy.classification.classification_achievable_performance_analysis <classification-achievable-performance-analysis>`
		"""
		if problem is None:
			problem = 'classification' if self.is_discrete(label_column) else 'regression'
		columns = [col for col in self._obj.columns if col != label_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		x_c = self._obj[continuous_columns].values if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values if len(discrete_columns) > 0 else None

		res = regression_achievable_performance_analysis(x_c, y, x_d=x_d, space=space, categorical_encoding=categorical_encoding) \
			if problem == 'regression' else classification_achievable_performance_analysis(x_c, y, x_d=x_d, space=space, \
				categorical_encoding=categorical_encoding)
		res = res.style.hide_index()

		return res


	def variable_selection_analysis(self, label_column, input_columns=(), space='dual', categorical_encoding='two-split', problem=None):
		"""
		Runs the model variable selection analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.

		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

				* :code:`'Variable'`: The column name corresponding to the input variable.
				* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
				* :code:`'Univariate Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal R^2 Increase'`: The highest amount by which the :math:`R^2` can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Univariate Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal Accuracy Increase'`: The highest amount by which the classification accuracy can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Conditional Mutual Information (nats)'`: The mutual information between this variable and the label, conditional on all variables previously selected.
				* :code:`'Running Mutual Information (nats)'`: The mutual information between all variables selected so far, including this one, and the label.
				* :code:`'Univariate Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model solely using this variable.
				* :code:`'Maximum Marginal True Log-Likelihood Per Sample Increase'`: The highest amount by which the true log-likelihood per sample can be increased as a result of adding this variable in the variable selection scheme.
				* :code:`'Running Achievable True Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a classification model using all variables selected so far, including this one.
				* :code:`'Maximum Marginal True Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase as a result of adding this variable.
				* :code:`'Running Maximum Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase (over the log-likelihood of the naive strategy consisting of predicting the mode of :math:`y`) as a result of using all variables selected so far, including this one.


		.. admonition:: Theoretical Foundation

			Section :ref:`2 - Variable Selection Analysis`.


		.. seealso::

			* :ref:`kxy.regression.regression_variable_selection_analysis <regression-variable-selection-analysis>`
			* :ref:`kxy.classification.classification_variable_selection_analysis <classification-variable-selection-analysis>`
		"""
		if problem is None:
			problem = 'classification' if self.is_discrete(label_column) else 'regression'

		columns = [col for col in self._obj.columns if col != label_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_variable_selection_analysis(x_c, y, x_d=x_d, space=space, categorical_encoding=categorical_encoding) \
			if problem == 'regression' else classification_variable_selection_analysis(x_c, y, x_d=x_d, space=space, \
				categorical_encoding=categorical_encoding)

		variable_columns = continuous_columns + discrete_columns
		res['Variable'] = res['Variable'].map({i: variable_columns[i] for i in range(len(variable_columns))})
		res.set_index(['Variable'], inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		res = res.style.background_gradient(cmap=cm)

		return res


	def auto_predictability(self, columns=(), space='primal', robust=True, p=None):
		"""
		Estimates the measure of auto-predictability of the time series corresponding to the input columns.

		.. math::

			\\mathbb{PR}\\left(\\{x_t \\} \\right)		:&= h\\left(x_* \\right) - h\\left( \\{ x_t \\} \\right) \\

														 &= h\\left(u_{x_*}\\right) - h\\left( \\{ u_{x_t} \\} \\right).


		Parameters
		----------
		columns : list
			The input columns.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			When p is None (the default), it is inferred from the sample.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.


		Returns
		-------
		 : float
			The measure of auto-predictability in nats/period.


		.. note::
			The time series is assumed to have a fixed period, and the index is assumed to be sorted.


		.. seealso::

			* :ref:`kxy.api.core.entropy_rate.auto_predictability <auto-predictability>`
		"""
		if columns == ():
			columns = self._obj.columns
		aps = [(col, auto_predictability(self._obj[col].values, space=space, robust=robust, p=p)) for col in columns]
		aps = sorted(aps, key=lambda x: x[1])
		res = pd.DataFrame({_[0]: [_[1]] for _ in aps})

		return res.T.rename(columns={0: 'PR'})




	def dataset_valuation(self, label_column, existing_input_columns, new_input_columns, space="dual", categorical_encoding="two-split", problem=None):
		"""
		Quantifies the additional performance that can be brought about by adding a set of new variables, as the difference between achievable performance with and without the new dataset.


		Parameters
		----------
		label_column : str
			The name of the column containing the true labels of the supervised learning problem.
		existing_input_columns : set
			List of columns corresponding to existing inputs/variables.
		new_input_columns : set
			List of columns corresponding to the new dataset.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.

		Returns
		-------
		res : pandas.Styler
			res.data is a Dataframe with columns (where applicable):

			* :code:`'Increased Achievable R^2'`: The highest :math:`R^2` increase that can result from adding the new dataset.
			* :code:`'Increased Achievable Log-Likelihood Per Sample'`: The highest increase in the true log-likelihood per sample that can result from adding the new dataset.
			* :code:`'Increased Achievable Accuracy'`: The highest increase in the classification accuracy that can result from adding the new dataset.


		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.

		.. seealso::

			* :ref:`kxy.regression.regression_achievable_performance_analysis <regression-achievable-performance-analysis>`
			* :ref:`kxy.classification.classification_achievable_performance_analysis <classification-achievable-performance-analysis>`
		"""
		if existing_input_columns is None or len(existing_input_columns) == 0:
			return self.achievable_performance_analysis(label_column, input_columns=new_input_columns, \
				categorical_encoding=categorical_encoding)

		if new_input_columns is None or len(new_input_columns) == 0:
			if problem is None:
				problem = 'classification' if self.is_discrete(label_column) else 'regression'

			if problem == 'classification':
				probas = [1.*(self._obj[self._obj[label_column]==cat].shape[0])/self._obj.shape[0] for cat in list(set(list(self._obj[label_column].values)))]
				return pd.DataFrame({\
					'Achievable R^2': [0.0], \
					'Achievable Log-Likelihood Per Sample': [0.0], \
					'Achievable Accuracy': [np.max(probas)]})

			else:
				return pd.DataFrame({\
					'Achievable R^2': [0.0], \
					'Achievable Log-Likelihood Per Sample': [0.0]})	

		all_inputs = set(list(existing_input_columns)+list(new_input_columns))
		new_perf = self.achievable_performance_analysis(label_column, input_columns=all_inputs, space=space, \
			categorical_encoding=categorical_encoding)
		old_perf = self.achievable_performance_analysis(label_column, input_columns=existing_input_columns, space=space, \
			categorical_encoding=categorical_encoding)
		imp_perf = new_perf-old_perf
		imp_perf.rename(columns={col: col.replace('Achievable', 'Increased Achievable') for col in imp_perf.columns}, inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		imp_perf = imp_perf.style.background_gradient(cmap=cm)

		return imp_perf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Dr. Yves-Laurent Kom Samo"
__copyright__ = "Copyright (C) 2020 KXY Technologies, Inc."
__license__ = """
Copyright (C) 2020 KXY TECHNOLOGIES, INC.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__version__ = "0.3.3"

from kxy.api import *
from kxy.regression import *
from kxy.classification import *
from kxy.asset_management import *
from kxy.pandas_extension import *


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd


from kxy.api import mutual_information_analysis
from kxy.api.core import scalar_continuous_entropy, prepare_data_for_mutual_info_analysis


def regression_achievable_performance_analysis(x_c, y, x_d=None, space='dual', categorical_encoding='two-split'):
	"""
	.. _regression-achievable-performance-analysis:
	Quantifies the :math:`R^2` that can be achieved when trying to predict :math:`y` with :math:`x`.

	.. math::

		\\bar{R}^2 = 1 - e^{-2I(y; f(x))}


	Parameters
	----------
	x_c : (n,) or (n, d) np.array
		Continuous inputs.
	x_d : (n,) or (n, d) np.array or None (default), optional
		Discrete inputs.
	y : (n,) np.array
		Labels.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model using provided inputs.
		* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a regression model using provided inputs.
		* :code:`'Achievable RMSE'`: The lowest RMSE that can be achieved by a regression model using provided inputs.

	.. admonition:: Theoretical Foundation

		Section :ref:`1 - Achievable Performance`.
	"""
	data = prepare_data_for_mutual_info_analysis(x_c, x_d, y, None, space=space, \
		non_monotonic_extension=True, categorical_encoding=categorical_encoding)
	output_indices = data['output_indices']
	corr = data['corr']
	batch_indices = data['batch_indices']

	mi_analysis = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)
	mi = mi_analysis['mutual_information']
	hy = scalar_continuous_entropy(y, space=space)
	std_y = np.sqrt(np.var(y))

	return pd.DataFrame({\
		'Achievable R^2': [1.-np.exp(-2.*mi)], \
		'Achievable True Log-Likelihood Per Sample': [hy-mi],
		'Achievable RMSE': std_y*np.exp(-mi)})



def regression_variable_selection_analysis(x_c, y, x_d=None, space='dual', categorical_encoding='two-split'):
	"""
	.. _regression-variable-selection-analysis:
	Runs the variable selection analysis for a regression problem.


	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	y : (n,) np.array
		Labels.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Variable'`: The variable index starting from 0 at the leftmost column of :code:`x_c` and ending at the rightmost column of :code:`x_d`.
		* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
		* :code:`'Univariate Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model solely using this variable.
		* :code:`'Maximum Marginal R^2 Increase'`: The highest amount by which the :math:`R^2` can be increased as a result of adding this variable in the variable selection scheme.
		* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model using all variables selected so far, including this one.
		* :code:`'Conditional Mutual Information (nats)'`: The mutual information between this variable and the label, conditional on all variables previously selected.
		* :code:`'Running Mutual Information (nats)'`: The mutual information between all variables selected so far, including this one, and the label.
		* :code:`'Maximum Marginal True Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase as a result of adding this variable.
		* :code:`'Running Maximum Log-Likelihood Increase Per Sample'`: The highest amount by which the true log-likelihood per sample can increase (over the log-likelihood of the naive strategy consisting of predicting the mode of :math:`y`) as a result of using all variables selected so far, including this one.


	.. admonition:: Theoretical Foundation

		Section :ref:`2 - Variable Selection Analysis`.
	"""
	data = prepare_data_for_mutual_info_analysis(x_c, x_d, y, None, space=space, \
		non_monotonic_extension=True, categorical_encoding=categorical_encoding)
	output_indices = data['output_indices']
	corr = data['corr']
	batch_indices = data['batch_indices']

	mi_analysis = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)

	d = len(batch_indices)
	batches = [_ for _ in range(d)]
	remaining_columns = [_ for _ in range(d)]

	if mi_analysis is None:
		return None

	std_y = np.sqrt(np.var(y))
	cmis = {}
	rsqs = {}
	rmses = {}
	mis = {}
	order = {}
	idx = 1
	for i in range(1, d+1):
		column_id = mi_analysis['selection_order'][str(i)]
		column = batches[column_id]
		if column in remaining_columns:
			mis[column] = mi_analysis['individual_mutual_informations'][str(i)]
			if idx == 1:
				cmis[column] = mis[column]
			else:
				cmis[column] = mi_analysis['conditional_mutual_informations'][str(i)]
			rsqs[column] = 1.-np.exp(-2.*mis[column])
			rmses[column] = std_y*np.exp(-mis[column])
			order[column] = idx
			idx += 1
			remaining_columns.remove(column)

	rsq_inc = {}
	rmse_dec = {}
	run_mi = 0
	run_rsqs = {}
	run_rmses = {}
	for v, o in sorted(order.items(), key=lambda item: item[1]):
		run_mi += cmis[v]
		run_rsqs[o] = 1.-np.exp(-2.*run_mi)
		run_rmses[o] = std_y*np.exp(-run_mi)

		if rsq_inc == {}:
			rsq_inc[o] = 1.-np.exp(-2.*cmis[v])
			rmse_dec[o] = std_y-std_y*np.exp(-cmis[v])
			
		else:
			rsq_inc[o] = run_rsqs[o]-run_rsqs[o-1]
			rmse_dec[o] = run_rmses[o-1]-run_rmses[o]

	n = y.shape[0]
	max_r_2 = np.max([_ for _ in run_rsqs.values()])
	importance_df = pd.DataFrame({
		'Variable': [v for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Selection Order': [o for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# Achievable R^2 analysis
		'Univariate Achievable R^2': [rsqs[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Maximum Marginal R^2 Increase': [rsq_inc[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable R^2': [run_rsqs[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable R^2 (%)': [100.*run_rsqs[o]/max_r_2 for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# Achievable RMSE analysis
		'Univariate Achievable RMSE': [rmses[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Maximum Marginal RMSE Decrease': [rmse_dec[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable RMSE': [run_rmses[o] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Running Achievable RMSE/STD (%)': [100.*run_rmses[o]/std_y for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# Conditional mutual information analysis
		'Univariate Mutual Information (nats)': [mis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Conditional Mutual Information (nats)': [cmis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \

		# The largest likelihood by which the likelihood can be increased as a result of adding this variable
		'Univariate Maximum True Log-Likelihood Increase Per Sample': [mis[v] for v, o in sorted(order.items(), key=lambda item: item[1])], \
		'Maximum Marginal True Log-Likelihood Increase Per Sample': [cmis[v] for v, o in sorted(order.items(), key=lambda item: item[1])]})

	importance_df['Running Mutual Information (nats)'] = importance_df['Conditional Mutual Information (nats)'].cumsum()
	max_mi = importance_df['Running Mutual Information (nats)'].max()
	importance_df['Running Mutual Information (%)'] = 100.*importance_df['Running Mutual Information (nats)']/max_mi

	# The largest value by which the lof likelihood can be increased by using the variables selected so far.
	importance_df['Running Maximum Log-Likelihood Increase Per Sample'] = importance_df['Maximum Marginal True Log-Likelihood Increase Per Sample'].cumsum()

	return importance_df





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from .risk_analysis import information_adjusted_correlation


def information_adjusted_beta(r, r_m):
	"""
	.. _information-adjusted-beta:
	Calculates the informmation-adjusted beta.



	.. note::

		The standard beta coefficient of an asset or a portfolio is defined by the CAPM model

		.. math::

			r = \\alpha + r_f + \\beta (r_m-r_f) + \\epsilon

		where :math:`r_f` is a deterministic risk-free rate, :math:`r_m` represents market returns, and
		:math:`\\epsilon` an idiosyncratic noise term. It follows that

		.. math::
			\\beta = \\text{Corr}\\left(r, r_m\\right) \\sqrt{\\frac{\mathbb{V}\\text{ar}\\left(r\\right)}{\mathbb{V}\\text{ar}\\left(r_m\\right)}},

		where :math:`\\text{Corr}` is Pearson's correlation coefficient. 

		The information-adjusted correlation generalizes the foregoing equations and reads 

		.. math::
			\\text{IA}\\beta = \\text{IACorr}\\left(\\{r\\}, \\{r_m\\}\\right) \\sqrt{\\frac{\mathbb{V}\\text{ar}\\left(r\\right)}{\mathbb{V}\\text{ar}\left(r_m\\right)}}.

		While Pearson's correlation (and therefore beta) only captures linear relationships between portfolio returns and market returns, the information-adjusted correlation fully captures nonlinear and temporal dependencies between portfolio returns and market returns.


	.. seealso::

		:ref:`kxy.finance.risk_analysis.information_adjusted_correlation <information-adjusted-correlation>`


	Parameters
	----------
	r : (n,) or (n,d) np.array
		The array of asset(s) or portfolio(s) returns.
	r_m : (n,) np.array
		The array of market returns.


	Returns
	-------
	c : float
		The information-adjusted of the asset or portfolio.

	Raises
	------
	AssertionError
		If any returns array is not one-dimensional.
	"""
	assert len(r_m.shape)==1 or r_m.shape[1]==1, 'r_m should be one-dimensional'

	return information_adjusted_correlation(r_m, r).flatten() * np.nansqrt(np.nanvar(r, axis=0)/np.nanvar(r_m))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from kxy.api import mutual_information_analysis
from kxy.api.core import prepare_data_for_mutual_info_analysis, spearman_corr, pearson_corr

from kxy.asset_management import information_adjusted_correlation, robust_pearson_corr


class BaseAccessor(object):
	"""
	Base class inheritated by our customs accessors.
	"""
	def __init__(self, pandas_obj):
		self._obj = pandas_obj


	def corr(self, columns=(), method='information-adjusted', min_periods=1, p=0, p_ic='hqic'):
		"""
		Calculates the auto-correlation matrix of all columns or the input subset.


		Parameters
		----------
		columns : set, optional
			The set of columns to use. If not provided, all columns are used.
		method : str, optional
			Which method to use to calculate the auto-correlation matrix. Supported
			values are 'information-adjusted' (the default) and all 'method' values of pandas.DataFrame.corr.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			The default value is 0, which corresponds to assuming rows are i.i.d. Values other than 0 are only
			supported in the robust-pearson method. When p is None, it is inferred from the sample.
		min_periods : int, optional
			Only used when method is not 'information-adjusted'. 
			See the documentation of pandas.DataFrame.corr.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.



		Returns
		-------
		c : pandas.DataFrame
			The auto-correlation matrix.


		.. seealso::

			:ref:`kxy.finance.risk_analysis.information_adjusted_correlation <information-adjusted-correlation>`
			:ref:`kxy.finance.risk_analysis.robust_pearson_corr <robust-pearson-corr>`
		"""
		columns = self._obj.columns if columns == () else list(columns)

		if method == 'information-adjusted':
			c = information_adjusted_correlation(self._obj[columns].values, y=None)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'robust-pearson':
			c = robust_pearson_corr(self._obj[columns].values, y=None, p=p, p_ic=p_ic)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'spearman':
			c = spearman_corr(self._obj[columns].values)
			return pd.DataFrame(c, columns=columns, index=columns)

		if method == 'pearson':
			c = pearson_corr(self._obj[columns].values)
			return pd.DataFrame(c, columns=columns, index=columns)

		else:
			return pd.DataFrame.corr(self._obj[columns], method=method, min_periods=min_periods)


	def is_discrete(self, column):
		"""
		Determine whether the input column contains discrete observations.
		"""
		ret = (not np.can_cast(self._obj[column].values, float))
		ret = ret or len(list(set(self._obj[column].values))) < 0.5*self._obj.shape[0]

		return ret


	def is_categorical(self, column):
		"""
		Determine whether the input column contains categorical observations.
		"""
		ret = (not np.can_cast(self._obj[column].values, float))

		return ret


	def _mutual_information_analysis(self, label_column, input_columns=(), space='dual', \
		categorical_encoding="two-split", non_monotonic_extension=True):
		"""
		"""
		columns = input_columns if len(input_columns) > 0 else [_ for _ in self._obj.columns if _ != label_column]
		cont_columns =  [col for col in columns if not self.is_categorical(col)]
		cat_columns = [col for col in columns if self.is_categorical(col)]

		x_c = self._obj[cont_columns].values if len(cont_columns) > 0 else None
		x_d = self._obj[cat_columns].values if len(cat_columns) > 0 else None
		y_c = None if self.is_categorical(label_column) else self._obj[label_column].values
		y_d = None if not self.is_categorical(label_column) else self._obj[label_column].values

		res = prepare_data_for_mutual_info_analysis(x_c, x_d, y_c, y_d, space=space, \
			non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)
		output_indices = res['output_indices']
		corr = res['corr']
		batch_indices = res['batch_indices']
		mi_ana = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)

		return mi_ana, res


	def describe(self,):
		for col in sorted(self._obj.columns):
			print('       ')
			print('---------' + '-'.join(['' for c in col]))
			print('Column: %s' % col)
			print('---------' + '-'.join(['' for c in col]))
			if self._obj.kxy.is_categorical(col):
				print('Type:      Categorical')
				labels, counts = np.unique(self._obj[col].values.astype(str), return_counts=True)
				labels_with_counts = [(labels[i], 100.*counts[i]/self._obj.shape[0]) \
									  for i in range(len(labels))]
				labels_with_counts = sorted(labels_with_counts, key=lambda x: -x[1])
				tot = 0.0
				for label, freq in labels_with_counts:
					print('Frequency: %s%%, Label: %s' % (('%.2f' % freq).rjust(5, ' '), label))
					tot += freq
					if tot > 90. and tot < 100.:
						print('Other Labels: %.2f%%' % (100.-tot))
						break
			else:
				print('Type:   Continuous')
				print('Max:    %.4f' % self._obj[col].max())
				print('Mean:   %.4f' % self._obj[col].mean())
				print('Median: %.4f' % self._obj[col].median())
				print('Min:    %.4f' % self._obj[col].min())  


	def __hash__(self):
		return hash(self._obj.to_string())






import hashlib
import json

import boto3

import pandas as pd

from db import get_kxy_db_engine

if __name__ == '__main__':
	# boto3.setup_default_session(profile_name='kxy')
	# s3 = boto3.resource('s3')
	# bucket = 'api.kxysolutions.com'

	# engine = get_kxy_db_engine()
	# entropies = pd.read_sql('select * from rv_copula_entropy_all where d =2 limit 1;', engine)
	# entropies = entropies.to_dict(orient='records')

	# for e in entropies:
	# 	ordered_top_corr = e['ordered_spearman_corr_matrix']
	# 	d = e['d']
	# 	key = json.dumps(ordered_top_corr)
	# 	md5_key = hashlib.md5(str(key).encode()).hexdigest()
	# 	key = 'core/dependence/copula/maximum-entropy/entropy/all/d=%d/%s.json' % (d, md5_key)
	# 	print(ordered_top_corr, key)
	# 	# obj = s3.Object(bucket, key)
	# 	# data = {'mean_entropy': e['mean_entropy'], 'median_entropy': e['median_entropy'], \
	# 	# 	'count': e['count'], 'entropy_std_err': e['entropy_std_err'], \
	# 	# 	'ordered_spearman_corr_matrix': ordered_top_corr}
	# 	# obj.put(Body=json.dumps(data))


	# ordered_corr = {0: ['1.0', '0.100'], 1: ['0.100', '1.0']}

	# from db import get_session_factory, CopulaEntropyModel
	# session_factory = get_session_factory()
	# db_session = session_factory()
	# entropy = db_session.query(CopulaEntropyModel).filter_by(\
	# 	ordered_spearman_corr_matrix=ordered_corr).first()
	# if entropy:
	# 	print(float(entropy.median_entropy))


	from db import get_kxy_db_engine

	engine = get_kxy_db_engine()
	engine.execute('refresh materialized view rv_copula_entropy_all;')



import logging
from multiprocessing import cpu_count

from scipy.stats import ortho_group
import numpy as np
import pandas as pd

from workers import CopulaFinderWorkerPool
from db import get_kxy_db_engine


if __name__ == '__main__':
	logging.basicConfig(format='%(asctime)s - %(process)d - %(levelname)s - %(message)s', level=logging.INFO)
	logging.info('Creating the Queues')
	n_finders = cpu_count()

	logging.info('Generating the parameters')
	logging.info('Adding the parameters to the processing queue')

	worker_pool = None
	for d in range(2, 11):
		for i in range(10000):
			u = ortho_group.rvs(d)
			s = np.random.gamma(.7, 2., d)
			cov = np.dot(u.T, np.dot(np.diag(s), u))
			sd = np.sqrt(np.diag(cov))
			corr = ((cov / sd).T / sd).T

			if worker_pool is None:
				worker_pool = CopulaFinderWorkerPool(workers=cpu_count(), class_name='FullCopulaFinder')

			worker_pool.submit(corr)

	# Don't exit until we are done
	logging.info('Waiting for the work to be done')
	worker_pool.queue.join()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import os
import requests
from time import sleep, time

import numpy as np

from .client import APIClient


def mutual_information_analysis(corr, output_indices, space='dual', batch_indices=[]):
	'''
	Analyzes the dependency between :math:`d`-dimensional continuous random vector :math:`x=\\left(x_1, \\dots, x_d \\right)` and
	one or more continuous random scalar :math:`y`.


	Recall that, for any permutation :math:`\\pi_1, \\dots, \\pi_d` of :math:`1, \\dots, d`, by the tower law,

	.. math::

		I\\left(y; x_1, \\dots, x_d\\right) = I\\left(y; x_{\\pi_1}\\right) + \\sum_{i=2}^d I\\left(y; x_{\\pi_i} \\vert x_{\\pi_{i-1}}, \\dots, x_{\\pi_1} \\right).


	This function estimates the mutual information :math:`I(y; x)` by learning the following permutation. 

	* :math:`x_{\\pi_1}` is the input with the largest maximum entropy mutual information with :math:`y` under Spearman rank correlation constraints.

	* :math:`x_{\\pi_i}` for :math:`i>1` is the input with the largest maximum entropy conditional mutual information :math:`I\\left(y; * \\vert x_{\\pi_{i-1}}, \\dots, x_{\\pi_1}\\right)`. Note that by the time :math:`\\pi_i` is selected, :math:`I\\left(y; x_{\\pi_{i-1}}, \\dots, x_{\\pi_1}\\right)` is already known, so that the maximum entropy conditional mutual information is simply derived from the maximum entropy copula distribution of :math:`I\\left(y; x_{\\pi_i}, \\dots, x_{\\pi_1}\\right)`.

	This function returns the learned permutation of inputs, the associated conditional mutual informations (a.k.a, the incremental input importance scores), as well as the mutual information :math:`I\\left(y; x_1, \\dots, x_d\\right)`.


	Parameters
	----------
	corr : (d, d) np.array
		The Spearman correlation matrix.

	output_indices: list of int 
		The index (indices) of the column to use as output.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.



	Returns
	-------
	res : dict
		Dictionary with keys :code:`mutual_information`, :code:`selection_order`, and :code:`conditional_mutual_informations`.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	bi = json.dumps(batch_indices)
	oi = json.dumps(output_indices)

	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying mutual information analysis with corr=%s and output_indices=%s' % (c, output_indices))
			api_response = APIClient.route(path='/rv/mutual-information-analysis', method='POST',\
				corr=c, output_indices=oi, request_id=request_id, timestamp=int(time()), \
				space=space, batch_indices=bi)

			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying mutual information analysis for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/mutual-information-analysis', method='POST',\
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)

	if api_response.status_code == requests.codes.ok:
		return api_response.json()

	else:
		logging.warning(api_response.json())

	return None



def copula_entropy_analysis(corr, space='dual'):
	'''
	Analyzes the entropy of the copula of a :math:`d`-dimensional continuous random vector :math:`x=\\left(x_1, \\dots, x_d \\right)`, with copula-uniform representation :math:`u=\\left(u_1, \\dots, u_d \\right)`.

	Recall that, for any permutation :math:`(1), \\dots, (d)` of :math:`1, \\dots, d`, by the tower law,

	.. math::

		h\\left(u_1, \\dots, u_d\\right) = \\sum_{i=2}^d h\\left( u_{(i)} \\vert u_{(i-1)}, \\dots, u_{(1)} \\right).


	This function estimates the copula entropy  :math:`h(u)` by learning the following permutation:

	* :math:`x_{(1)}` and :math:`x_{(2)}` are chosen to be the two random variables with smallest copula entropy (or equivalently, the highest mutual information).

	* :math:`x_{(i)}` for :math:`i>1` is the input with the smallest conditional copula entropy :math:`h\\left(* \\vert u_{(i-1)}, \\dots, u_{(1)} \\right)` (or equivalently, the highest mutual information :math:`I\\left(*; x_{(1)}, \\dots, x_{(i-1)}\\right)`). Note that by the time :math:`(i)` is selected, :math:`h\\left(u_{(i-1)}, \\dots, u_{(1)}\\right)` is already known, so that the maximum entropy conditional entropy is simply derived from the maximum entropy copula distribution of :math:`\\left(x_{(i)}, \\dots, x_{(1)}\\right)`.

	This function returns the learned permutation of inputs, the association conditional entropies, as well as the copula entropy :math:`h(u)`.


	Parameters
	----------
	corr : np.array
		The Spearman correlation matrix.

	output_index: int
		The index of the column to use as output.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	res : dict
		Dictionary with keys :code:`copula_entropy`, :code:`selection_order`, and :code:`conditional_copula_entropies`.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying copula entropy analysis with corr=%s' % (c))
			api_response = APIClient.route(path='/rv/copula-entropy-analysis', method='POST', corr=c, \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying copula entropy analysis for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/copula-entropy-analysis', method='POST', \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)


	if api_response.status_code == requests.codes.ok:
		return api_response.json()

	else:
		logging.warning(api_response)
		logging.warning(api_response.json())

	return None




def information_adjusted_correlation_from_spearman(corr, space='dual'):
	'''
	Determine the Pearson correlation matrix that should be used as plug-in replacement
	for the standard Pearson correlation estimator, in situations where Gaussianity has
	to be assumed for simplicity, but the data-generating distribution exhibits non-Gaussian
	traits (e.g. heavier or lighter tails).

	We proceed as follows, for every pairwise Spearman rank correlation, we estmate the smallest
	mutual information that is consistent with the pairwise Spearman rank correlation. This is also 
	the mutual information of the model which, among all models that have the same Spearman rank 
	correlation, is the most uncertain about everything else. 

	This mutual information captures the essence of the dependence between the two variables without
	assuming Gaussianity. To go back to the Gaussian scale, we determine the Pearson correlation which, 
	under the Gaussian assumption would yield the same mutual information as estimated.


	Parameters
	----------
	corr : np.array
		The Spearman correlation matrix.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	res : np.array
		The array of equivalent Pearson correlation coefficients.
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''
	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Querying information adjusted correlation with corr=%s' % (c))
			api_response = APIClient.route(path='/rv/information-adjusted-correlation', method='POST', corr=c, \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Querying information adjusted correlation for request_id=%s' % request_id)
			api_response = APIClient.route(path='/rv/information-adjusted-correlation', method='POST', \
				request_id=request_id, timestamp=int(time()), space=space)
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)


	if api_response.status_code == requests.codes.ok:
		return np.array(api_response.json()['ia_corr']).astype(float)

	else:
		logging.warning(api_response.json())

	return None



def robust_pearson_corr_from_spearman(corr):
	'''
	Return the Pearson correlation matrix that is equivalent to the input 
	Spearman rank correlation matrix, assuming inputs are jointly Gaussian.


	Parameters
	----------
	corr : np.array
		The Spearman correlation matrix.


	Returns
	-------
	res : np.array
		The array of equivalent Pearson correlation coefficients.
	'''
	return information_adjusted_correlation_from_spearman(corr, space='primal')




def predict_copula_uniform(u_x_dict_list, corr, output_indices, space='dual', batch_indices=[], \
		problem_type='regression'):
	'''
	'''
	c = json.dumps([['%.3f' % corr[i, j]  for j in range(corr.shape[1])] for i in range(corr.shape[0])])
	ux = json.dumps(u_x_dict_list)
	bi = json.dumps(batch_indices)
	oi = json.dumps(output_indices)

	max_retry = 60
	first_try = True
	retry_count = 0
	request_id = ''

	while (first_try or api_response.status_code == requests.codes.retry) and retry_count < max_retry:
		first_try = False

		if request_id == '':
			query_start_time = time()
			# First attempt
			logging.debug('Requesting predictions endpoint.')
			api_response = APIClient.route(path='/rv/predictions', method='POST', request_type='auth',
			 timestamp=int(time()))
			request_id   = api_response.json()['request_id']
			response = api_response.json()['prediction_endpoint']
			prediction_endpoint = response['url']
			additional_fields = response['fields']

			logging.debug('Submitting prediction job.')
			params = {'corr': c, 'output_indices': oi, 'request_id': request_id, 'timestamp': int(time()), \
				'space': space, 'batch_indices': bi, 'u_x_dict_list': ux, 'problem_type': problem_type
			}

			key = 'rv_predict_full/' + request_id + '.json'
			os.makedirs('rv_predict_full', exist_ok=True)
			with open(key, 'w') as f:
				json.dump(params, f)

			with open(key, 'rb') as f:
				files = {'file': (key, f)}
				response = requests.post(prediction_endpoint, data=response['fields'], \
					files=files)

			logging.debug('Done submitting prediction job.')
			query_duration = time()-query_start_time

		else:
			query_start_time = time()
			# Subsequent attempt: refer to the initial request
			logging.debug('Requesting predictions.')
			api_response = APIClient.route(path='/rv/predictions', method='POST',\
				request_id=request_id, timestamp=int(time()))
			query_duration = time()-query_start_time

		retry_count += 1
		if api_response.status_code == requests.codes.retry:
			request_id = api_response.json()['request_id']
			sleep(.1 if query_duration > 10. else 10.)

	if api_response.status_code == requests.codes.ok:
		return api_response.json()

	else:
		logging.warning(api_response.json())

	return None






import numpy as np

from kxy.api import mutual_information_analysis, predict_copula_uniform
from kxy.api.core import prepare_data_for_mutual_info_analysis, prepare_test_data_for_prediction

from .pre_learning import regression_achievable_performance_analysis


class MaxEntRegressor(object):
	'''
	Regression model based on the maximum-entropy principle.
	'''

	def fit(self, x_c, y, x_d=None, space="dual", categorical_encoding="two-split"):
		"""
		Solves the maximum-entropy problem on the API, in the background. 

		The resulting joint copula-uniform distribution :math:`(u_y, u_x)` is the cornerstone of inference.


		Parameters
		----------
		x_c : (n,) or (n, d) np.array
			Continuous inputs.
		x_d : (n,) or (n, d) np.array or None (default), optional
			Discrete inputs.
		y : (n,) np.array
			Labels.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.


		Returns
		-------
		a : pandas.DataFrame

			Dataframe with columns:

			* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a regression model using provided inputs.
			* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a regression model using provided inputs.
		"""
		data = prepare_data_for_mutual_info_analysis(x_c, x_d, y, None, space=space, \
			non_monotonic_extension=True, categorical_encoding=categorical_encoding)
		output_indices = data['output_indices']
		corr = data['corr']
		batch_indices = data['batch_indices']
		self.corr_train = corr
		self.train_x_c = x_c.astype(float)
		self.train_x_d = x_d.astype(str) if x_d is not None else None
		self.train_y_c = y.astype(float).flatten()

		self.space = space
		self.categorical_encoding = categorical_encoding
		mi_analysis = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)

		return self.achievable_performance


	@property	
	def achievable_performance(self):
		"""
		Dataframe containing the highest performance that can be achieved. Requires the model to be fitted first.

		.. seealso::

			:ref:`kxy.regression.pre_learning.regression_achievable_performance_analysis <regression-achievable-performance-analysis>`

		"""
		assert getattr(self, 'train_x_c', None) is not None or getattr(self, 'train_x_d', None) is not None,\
			'The model has not been fitted yet.'

		return regression_achievable_performance_analysis(self.train_x_c, self.train_y_c, x_d=self.train_x_d, \
			space=self.space)


	def predict(self, x_c, x_d=None):
		"""
		Calculates the posterior mean and posterior standard deviation of the copula-uniform representation of the output

		.. math::
			E(u_y|x= *) \\text{ and } \\sqrt{Var \\left( u_y \\vert x= *  \\right)}

		under the maximum-entropy distribution for the copula-uniform representations :math:`(u_y, u_x)`, and infer predicted outputs.

		Missing inputs are handled gracefully, and the posterior distribution is based on provided inputs.


		Parameters
		----------
		x_c : (n,d) np.array
			Test continuous inputs. Missing inputs, if any, should be represented as np.nan or None.
		x_d : (n, d) np.array or None (default), optional
			Test discrete inputs. Missing inputs, if any, should be represented as np.nan or None.


		Returns
		-------
		 : dict

			Dictionary with keys:

				* :code:`posterior_mean_u_y`: The ndarray of the posterior mean of outputs encodings.
				* :code:`posterior_std_u_y`: The ndarray of the posterior std of outputs encodings.
				* :code:`predicted_labels`: The ndarray of predicted outputs corresponding to `posterior_mean_u_y`.
		"""
		test_x_c = x_c
		test_x_d = x_d

		res = prepare_test_data_for_prediction(test_x_c, test_x_d, self.train_x_c, self.train_x_d, \
			self.train_y_c, None, non_monotonic_extension=True, categorical_encoding=self.categorical_encoding, \
			space=self.space)

		corr = res['corr']
		u_x_dict_list  = res['u_x_dict_list']
		output_indices = res['output_indices']
		batch_indices  = res['batch_indices']
		problem_type   = res['problem_type']

		predictions = predict_copula_uniform(u_x_dict_list, corr, output_indices, space=self.space, \
			batch_indices=batch_indices, problem_type=problem_type)

		posterior_mean_u_y = np.array(predictions['posterior_means']).astype(float)
		np.copyto(posterior_mean_u_y, 0.5, where=np.isnan(posterior_mean_u_y))
		posterior_std_u_y  = np.array(predictions['posterior_stds']).astype(float)
		np.copyto(posterior_std_u_y, np.sqrt(1./12.), where=np.isnan(posterior_std_u_y))

		predicted_labels   = np.percentile(self.train_y_c, 100.0*posterior_mean_u_y.flatten(), axis=0)

		return {'posterior_mean_u_y': posterior_mean_u_y, 'posterior_std_u_y': posterior_std_u_y, 'predicted_labels': predicted_labels}









#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base_accessor import BaseAccessor

class LearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for automatically training predictive models.

	This class defines the :code:`kxy_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect

def decorate_methods(decorator, include=[]):
	"""
	Systematically decorates all methods of a class, except for the ones 
	listed in exclude.
	"""
	def decorate(cls):
		method_names = [_[0] for _ in inspect.getmembers(cls, inspect.isroutine) if _[0] in include]
		for attr in method_names:
			setattr(cls, attr, decorator(getattr(cls, attr)))
		return cls
	return decorate



def decorate_all_methods(decorator):
	"""
	Systematically decorates all methods of a class, except for the ones 
	listed in exclude.
	"""
	def decorate(cls):
		method_names = [_[0] for _ in inspect.getmembers(cls, inspect.isroutine)]
		for attr in method_names:
			setattr(cls, attr, decorator(getattr(cls, attr)))
		return cls
	return decorate


import gc
import logging
import time
from multiprocessing import Process, Manager
from threading import Thread

import numpy as np

from optimizers import FullSpearmanMaxEntCopulaOptimizer


class FullCopulaFinder(Process):
	''' 
	The role of a finder is to consume model parameters from a queue.
	'''
	def __init__(self, parameters_queue):
		logging.info("Creating a copula finder worker.")
		super(FullCopulaFinder, self).__init__()
		self.daemon = True
		self.parameters_queue = parameters_queue
		self.d = None

	def run(self):
		keep_trying = True
		logging.info("Starting to find copulas...")
		while keep_trying:
			while not self.parameters_queue.empty():
				keep_trying = False
				corr = self.parameters_queue.get()
				d = corr.shape[0]

				if self.d is None:
					self.d = d

				if self.d != d:
					logging.info('Received a different n_var: terminating process')
					self.parameters_queue.put(corr)
					self.terminate()

				logging.info('Processing parameter batch %s' % str(corr))
				m = FullSpearmanMaxEntCopulaOptimizer(spearman_correlation_matrix=corr, \
					record_copulas=True, optimize=True)

				self.parameters_queue.task_done()
				gc.collect()

			time.sleep(0.01)



class CopulaFinderWorkerPool(object):
    """
    Class for worker pool with builtin queue
    """

    def __init__(self, workers=15, class_name='FullCopulaFinder'):
        self.num_workers = workers
        self.class_name = class_name
        m = Manager()
        self.queue = m.Queue()
        self.workers = [self._init_worker() for _ in range(self.num_workers)]
        self.watcher_thread = Thread(target=self._watch_workers)
        self.watcher_thread.start()


    def _init_worker(self):
        worker = eval(self.class_name)(self.queue)
        worker.start()
        return worker


    def _watch_workers(self):
        while True:
            time.sleep(2.)
            logging.info('Queue Size: %s' % self.queue.qsize())
            for idx in range(self.num_workers):
                if idx > len(self.workers) - 1:
                    logging.warning("Adding a new worker...")
                    self.workers += [self._init_worker()]
                elif not self.workers[idx].is_alive():
                    logging.warning("Dead worker detected, Adding a new worker...")
                    self.workers[idx] = self._init_worker()

    def submit(self, args):
        """
        Adds the named params dict as a work item to worker queue
        :param kwargs: all the named params for this work item
        """
        self.queue.put(args, block=False)







#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import seaborn as sns

from kxy.api.core import auto_predictability, prepare_data_for_mutual_info_analysis

from kxy.classification import classification_model_improvability_analysis, \
	classification_model_explanation_analysis, classification_bias

from kxy.regression import regression_model_improvability_analysis, \
	regression_model_explanation_analysis, regression_bias

from .base_accessor import BaseAccessor


@pd.api.extensions.register_dataframe_accessor("kxy_post_learning")
class PostLearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for **post-learning** in supervised learning problems.

	This class defines the :code:`kxy_post_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_post_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""

	def model_improvability_analysis(self, label_column, model_prediction_column, input_columns=(), 
			space='dual', categorical_encoding='two-split', problem=None):
		"""
		Runs the model improvability analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`label_column` is categorical.


		Parameters
		----------
		label_column : str
			The name of the column containing true labels.
		model_prediction_column : str
			The name of the column containing labels predicted by the model.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`model_prediction_column` and :code:`label_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.


		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns (where applicable):

				* :code:`'Leftover R^2'`: The amount by which the trained model's :math:`R^2` can still be increased without resorting to additional inputs, simply through better modeling.
				* :code:`'Leftover Log-Likelihood Per Sample'`: The amount by which the trained model's true log-likelihood per sample can still be increased without resorting to additional inputs, simply through better modeling.
				* :code:`'Leftover Accuracy'`: The amount by which the trained model's classification accuracy can still be increased without resorting to additional inputs, simply through better modeling.


		.. admonition:: Theoretical Foundation

			Section :ref:`3 - Model Improvability`.

		.. seealso::

			* :ref:`kxy.regression.regression_model_improvability_analysis <regression-model-improvability-analysis>`
			* :ref:`kxy.classification.classification_model_improvability_analysis <classification-model-improvability-analysis>`
		"""
		if problem is None:
			problem = 'classification' if self.is_discrete(label_column) else 'regression'

		columns = [col for col in self._obj.columns if col != label_column and col != model_prediction_column] \
			if len(input_columns) == 0 else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		y = self._obj[label_column].values
		y_p = self._obj[model_prediction_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_model_improvability_analysis(x_c, y_p, y, x_d=x_d, space=space, \
				categorical_encoding=categorical_encoding) if problem == 'regression' \
			else classification_model_improvability_analysis(x_c, y_p, y, x_d=x_d, space=space, \
				categorical_encoding=categorical_encoding)
		res = res.style.hide_index()

		return res



	def model_explanation_analysis(self, model_prediction_column, input_columns=(), space='dual', categorical_encoding='two-split', problem=None):
		"""
		Runs the model explanation analysis on a trained supervised learning model.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`model_prediction_column` is categorical.

		Parameters
		----------
		model_prediction_column : str
			The name of the column containing predicted labels.
		input_columns : set
			List of columns to use as inputs. When an empty set/list is provided, all columns but :code:`model_prediction_column` are used as inputs.
		space : str, 'primal' | 'dual'
			The space in which the maximum entropy problem is solved. 
			When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
			When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.


		Returns
		-------
		res : pandas.Styler
			res.data is a pandas.Dataframe with columns:

				* :code:`'Variable'`: The column name corresponding to the input variable.
				* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
				* :code:`'Univariate Explained R^2'`: The :math:`R^2` between predicted labels and this variable.
				* :code:`'Running Explained R^2'`: The :math:`R^2` between predicted labels and all variables selected so far, including this one.
				* :code:`'Marginal Explained R^2'`: The increase in :math:`R^2` between predicted labels and all variables selected so far that is due to adding this variable in the selection scheme.



		.. admonition:: Theoretical Foundation

			Section :ref:`a) Model Explanation`.

		.. seealso::

			* :ref:`kxy.regression.regression_model_explanation_analysis <regression-model-explanation-analysis>`
			* :ref:`kxy.classification.classification_model_explanation_analysis <classification-model-explanation-analysis>`
		"""
		if problem is None:
			problem = 'classification' if self.is_discrete(model_prediction_column) else 'regression'

		columns = [col for col in self._obj.columns if col != model_prediction_column] if len(input_columns) == 0\
			else input_columns
		discrete_columns = [col for col in columns if self.is_categorical(col)]
		continuous_columns =  [col for col in columns if not self.is_categorical(col)]

		f_x = self._obj[model_prediction_column].values
		x_c = self._obj[continuous_columns].values.astype(float) if len(continuous_columns) > 0 else None
		x_d = self._obj[discrete_columns].values.astype(str) if len(discrete_columns) > 0 else None

		res = regression_model_explanation_analysis(x_c, f_x, x_d=x_d, space=space, categorical_encoding=categorical_encoding) if problem == 'regression' \
			else classification_model_explanation_analysis(x_c, f_x, x_d=x_d, space=space, categorical_encoding=categorical_encoding)

		variable_columns = continuous_columns + discrete_columns
		res['Variable'] = res['Variable'].map({i: variable_columns[i] for i in range(len(variable_columns))})
		res.set_index(['Variable'], inplace=True)

		cm = sns.light_palette("green", as_cmap=True)
		res = res.style.background_gradient(cmap=cm)

		return res



	def bias(self, bias_source_column, model_prediction_column, linear_scale=True, categorical_encoding='two-split', problem=None):
		"""
		Quantifies the bias in a supervised learning model as the mutual information between a possible cause and model predictions.

		The nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`model_prediction_column` is categorical.


		Parameters
		----------
		bias_source_column : str
			The name of the column containing values of the bias factor (e.g. age, gender, etc.)
		model_prediction_column : str
			The name of the column containing predicted labels associated to the values of :code:`bias_source_column`.
		linear_scale : bool
			Whether the bias should be returned in the linear/correlation scale or in the mutual information scale (in nats).
		categorical_encoding : str, 'one-hot' | 'two-split' (default)
			The encoding method to use to represent categorical variables. 
			See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.
		problem : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.

		Returns
		-------
		 : float
			The mutual information :math:`m` or :math:`1-e^{-2m}` if :code:`linear_scale=True`.


		.. admonition:: Theoretical Foundation

			Section :ref:`b) Quantifying Bias in Models`.

		.. seealso::

			* :ref:`kxy.regression.regression_bias <regression-bias>`
			* :ref:`kxy.classification.classification_bias <classification-bias>`
		"""
		if problem is None:
			problem = 'classification' if self.is_discrete(model_prediction_column) else 'regression'

		f_x = self._obj[model_prediction_column].values
		z = self._obj[bias_source_column].values

		return regression_bias(f_x, z, linear_scale=linear_scale, categorical_encoding=categorical_encoding) if problem == 'regression' \
			else classification_bias(f_x, z, linear_scale=linear_scale)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import requests

import numpy as np
import scipy.special as spe
import statsmodels.api as sm


from .utils import spearman_corr, pearson_corr


def _scalar_continuous_entropy(x, space='dual', method='gaussian-kde'):
	if space == 'primal' or method == 'gaussian':
		return 0.5*np.log(2.*np.pi*np.e*np.var(x))

	if method == '1-spacing':
		sorted_x = np.unique(x)
		n = sorted_x.shape[0]
		return np.sum(np.log(n*(sorted_x[1:]-sorted_x[:-1])))/n - spe.digamma(1)

	if method == 'gaussian-kde':
		kde = sm.nonparametric.KDEUnivariate(x)
		kde.fit(kernel='gau')
		return kde.entropy



def scalar_continuous_entropy(x, space='dual', method='gaussian-kde'):
	"""
	.. _scalar-continuous-entropy:
	Estimates the (differential) entropy of a continuous scalar random variable.

	Multiple methods are supported:

	* :code:`'gaussian'` for Gaussian moment matching: :math:`h(x) = \\frac{1}{2} \\log\\left(2 \\pi e \\sigma^2 \\right)`.
	* :code:`'1-spacing'` for the standard 1-spacing estimator (see [1]_ and [2]_):

	.. math::
		h(x) \\approx - \\gamma(1) + \\frac{1}{n-1} \\sum_{i=1}^{n-1} \\log \\left[ n \\left(x_{(i+1)} - x_{(i)} \\right) \\right],

	where :math:`x_{(i)}` is the i-th smallest sample, and :math:`\\gamma` is `the digamma function. <https://en.wikipedia.org/wiki/Digamma_function>`_
	
	* :code:`'gaussian-kde'` (the default) for Gaussian kernel density estimation.

	.. math::
		h(x) \\approx \\frac{1}{n} \\sum_{i=1}^n \\log\\left( \\hat{p}\\left(x_i\\right) \\right)

	where :math:`\\hat{p}` is the Gaussian kernel density estimator of the true pdf using :code:`statsmodels.api.nonparametric.KDEUnivariate`.

	
	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the continuous scalar random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape or the method is not supported.
		

	.. rubric:: References

	.. [1] Kozachenko, L. F., and Nikolai N. Leonenko. "Sample estimate of the entropy of a random vector." 
		Problemy Peredachi Informatsii 23.2 (1987): 9-16.

	.. [2] Beirlant, J., Dudewicz, E.J., Györfi, L., van der Meulen, E.C. "Nonparametric entropy estimation: an overview." 
		International Journal of Mathematical and Statistical Sciences. 6 (1): 17–40. (1997) ISSN 1055-7490. 
	"""
	assert len(x.shape) == 1 or x.shape[1] == 1, 'x should be a one dimensional numpy array'
	assert method in ('gaussian', '1-spacing', 'gaussian-kde')

	# Check if the distribution is a mixture of discrete and continuous
	categories, counts  = np.unique(x.flatten(), return_counts=True)
	n = x.shape[0]
	int_proba = 0.01
	exp_n_sample = n*int_proba
	std_n_sample = np.sqrt(n*(int_proba-int_proba**2))
	threshold = exp_n_sample + 5*std_n_sample

	discrete_part = categories[counts>threshold]
	continuous_part = categories[counts<=threshold]

	if len(discrete_part) == 0:
		return _scalar_continuous_entropy(x, space=space, method=method)


	if len(continuous_part) <= 20:
		return discrete_entropy(discrete_part)


	ent_disc_part = discrete_entropy(discrete_part)
	ent_cont_part = _scalar_continuous_entropy(continuous_part, space=space, method=method)
	proba_discrete = 1.*len(discrete_part)/n
	proba_continuous = 1.-proba_discrete 
	ent_disc_cont = -proba_discrete*np.log(proba_discrete)-proba_continuous*np.log(proba_continuous)

	ent = ent_disc_part + ent_cont_part - ent_disc_cont
	assert not np.isnan(ent)

	return ent



def discrete_entropy(x):
	"""
	.. _discrete-entropy:
	Estimates the (Shannon) entropy of a discrete random variable taking up to q distinct values 
	given n i.i.d samples,

	.. math::
		h(x) = - \\sum_{i=1}^q p_i \\log p_i,

	using the plug-in estimator.


	Parameters
	----------
	x : (n,) np.array
		i.i.d. samples from the distribution of interest.

	Returns
	-------
	h : float
		The (differential) entropy of the discrete random variable of interest.

	Raises
	------
	AssertionError
		If the input has the wrong shape.

	"""
	assert len(x.shape) == 1, 'Only flat numpy arrays are supported'
	categories = list(set(list(x)))
	n = len(x)
	probas = np.array([1.*len(x[x==cat])/n for cat in categories])

	return -np.dot(np.log(probas), probas)


def least_structured_copula_entropy(x, space='dual', batch_indices=[]):
	"""
	.. _least-structured-copula-entropy:
	Estimates the entropy of the maximum-entropy copula in the chosen space.

	.. note::
	
		This also corresponds to least amount of total correlation that is evidenced by our maximum-entropy constraints.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	h : float
		The (differential) entropy of the least structured copula consistent with maximum-entropy constraints.
	"""
	if len(x.shape) == 1 or x.shape[1] == 1:
		# By convention, the copula-dual representation of a 1d random variable is the uniform[0, 1].
		return 0.0

	x = x.astype(float)
	corr = pearson_corr(x) if space == 'primal' else spearman_corr(x)

	res = copula_entropy_analysis(corr, space=space)
	if res is None:
		return None
	h = res['copula_entropy']

	return h


def least_structured_continuous_entropy(x, space='dual', batch_indices=[]):
	""" 
	.. _least-structured-continuous-entropy:
	Estimates the entropy of a continuous :math:`d`-dimensional random variable under the least structured assumption for its copula. 

	When :math:`d>1`,

	.. math::
		h(x) = h(u) + \\sum_{i=1}^d h(x_i).

	* :math:`h(u)` is estimated using :ref:`least_structured_copula_entropy <least-structured-copula-entropy>`.
	* :math:`h(x_i)` are estimated using :ref:`scalar_continuous_entropy <scalar-continuous-entropy>`. 


	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.


	Returns
	-------
	h : float
		The (differential) entropy of the data generating distribution, assuming its copula is maximum-entropy in the chosen space. By convention, when :math:`d=1`, this function is the same as :ref:`scalar_continuous_entropy <scalar-continuous-entropy>`, and returns 0 when :math:`n=1`.
	"""
	if x.shape[0] == 1:
		# By convention, the entropy of a single sample is 0.
		return 0.0

	if len(x.shape) == 1 or x.shape[1] == 1:
		return scalar_continuous_entropy(x, space=space)

	ch = least_structured_copula_entropy(x, space=space, batch_indices=batch_indices)
	ih = np.sum([scalar_continuous_entropy(x[:, i], space=space) for i in range(x.shape[1])])

	return ih+ch




def least_structured_mixed_entropy(x_c, x_d, space='dual', batch_indices=[]):
	"""
	.. _least-structured-mixed-entropy:
	Estimates the joint entropy :math:`h(x_c, x_d)`, where :math:`x_c` is continuous random vector and :math:`x_d` is a discrete random vector.

	.. note::

		We use the identities

		.. math::
			h(x, y) &= h(y) + h(x|y) \\

			        &= h(y) + E \\left[ h(x \\vert y=.) \\right]

		that are true when :math:`x` and :math:`y` are either both continuous or both discrete
		to extend the definition of the joint entropy to the case where one is continuous and
		the other discrete.

		Specifically,

		.. math::
			h(x_c, x_d) = h(x_d) + \sum_{j=1}^q \mathbb{P}(x_d=j) h\\left(x_c \\vert x_d=j \\right).

		* :math:`h(x_d)` is estimated using :ref:`discrete_entropy <discrete-entropy>`.
		* :math:`\mathbb{P}(x_d=j)` is estimated using frequencies.
		* :math:`h\\left(x_c \\vert x_d=j \\right)` is estimated using :ref:`least_structured_continuous_entropy <least-structured-continuous-entropy>`.


	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous data generating distribution.

	x_d : (n,) np.array
		n i.i.d. draws from the discrete data generating distribution, jointly sampled with x_c.

	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	h : float
		The entropy of the least structured distribution as evidenced by average the maximum entropy constraints.
	"""
	labels = x_d.copy() if len(x_d.shape) == 1 else np.array([','.join([str(_) for _ in row]) for row in x_d])
	categories = list(set(labels))

	n = labels.shape[0]
	probas = np.array([1.*len(labels[labels==cat])/n for cat in categories])
	h = -np.dot(probas, np.log(probas))
	h += np.sum([probas[i] * least_structured_continuous_entropy(x_c[labels==categories[i]], space=space, \
		batch_indices=batch_indices) for i in range(len(categories))])

	return h






from _thread import RLock



_NOT_FOUND = object()
class cached_property(object):
	'''
	cached_property() - computed once per instance, cached as attribute
	'''
	def __init__(self, func):
		self.func = func
		self.attrname = None
		self.__doc__ = func.__doc__
		self.lock = RLock()

	def __set_name__(self, owner, name):
		if self.attrname is None:
			self.attrname = name
		elif name != self.attrname:
			raise TypeError(
				"Cannot assign the same cached_property to two different names "
				f"({self.attrname!r} and {name!r})."
			)

	def __get__(self, instance, owner=None):
		if instance is None:
			return self
		if self.attrname is None:
			raise TypeError(
				"Cannot use cached_property instance without calling __set_name__ on it.")
		try:
			cache = instance.__dict__
		except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
			msg = (
				f"No '__dict__' attribute on {type(instance).__name__!r} "
				f"instance to cache {self.attrname!r} property."
			)
			raise TypeError(msg) from None
		val = cache.get(self.attrname, _NOT_FOUND)
		if val is _NOT_FOUND:
			with self.lock:
				# check if another thread filled cache while we awaited lock
				val = cache.get(self.attrname, _NOT_FOUND)
				if val is _NOT_FOUND:
					val = self.func(instance)
					try:
						cache[self.attrname] = val
					except TypeError:
						msg = (
							f"The '__dict__' attribute on {type(instance).__name__!r} instance "
							f"does not support item assignment for caching {self.attrname!r} property."
						)
						raise TypeError(msg) from None
		return val



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np
from scipy.stats.mstats import rankdata
from numpy.dual import svd



def spearman_corr(x):
	"""
	Calculate the Spearman rank correlation matrix, ignoring nans.

	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable, whose Spearman rank correlation matrix this 
		function calculates.


	Returns
	-------
	corr : np.array
		The Spearman rank correlation matrix.
	"""
	mask = np.isnan(x).copy()
	valid_mask = np.logical_not(mask).astype(int)
	R = rankdata(x, axis=0)
	np.copyto(R, np.nan, where=mask)
	mean_R = np.nanmean(R, axis=0)
	demeaned_R = R - mean_R
	istd_R = 1./np.nanstd(demeaned_R, axis=0)
	np.copyto(istd_R, 0.0, where=np.isinf(istd_R))
	np.copyto(istd_R, 0.0, where=np.isnan(istd_R))
	standard_R = demeaned_R*istd_R
	np.copyto(standard_R, 0.0, where=mask)
	non_nan_ns = np.einsum('ij, ik->jk', valid_mask, valid_mask) # Number of non-nan pairs
	corr = np.divide(np.einsum('ij, ik->jk', standard_R, standard_R), non_nan_ns)
	np.fill_diagonal(corr, 1.)
	corr = np.round(corr, 3)

	return corr



def pearson_corr(x):
	"""
	Calculate the Pearson correlation matrix, ignoring nans.

	Parameters
	----------
	x : (n, d) np.array
		Input data representing n i.i.d. draws from the d-dimensional 
		random variable, whose Pearson correlation matrix this 
		function calculates.


	Returns
	-------
	corr : np.array
		The Pearson correlation matrix.
	"""
	n, m = x.shape
	mask = np.isnan(x).copy()
	valid_mask = np.logical_not(mask).astype(int)

	np.copyto(x, np.nan, where=mask)
	o = np.ones((m, m))
	non_nan_ns = np.einsum('ij, ik->jk', valid_mask, valid_mask) # Number of non-nan pairs
	mean_x = np.nanmean(x, axis=0)
	demeaned_x = x - mean_x
	istd_x = 1./np.nanstd(demeaned_x, axis=0)
	np.copyto(istd_x, 0.0, where=np.isinf(istd_x))
	np.copyto(istd_x, 0.0, where=np.isnan(istd_x))
	standard_x = demeaned_x*istd_x
	np.copyto(standard_x, 0.0, where=mask)
	corr = np.divide(np.einsum('ij, ik->jk', standard_x, standard_x), non_nan_ns)
	np.fill_diagonal(corr, 1.)
	corr = np.round(corr, 3)

	return corr




def robust_log_det(c):
	"""
	Computes the logarithm of the determinant of a positive definite matrix in a fashion that is more robust to ill-conditioning than taking the logarithm of np.linalg.det.

	.. note::
		Specifically, we compute the SVD of c, and return the sum of the log of eigenvalues. np.linalg.det on the other hand computes the Cholesky decomposition of c, which is more likely to fail than its SVD, and takes the product of its diagonal elements, which could be subject to underflow error when diagonal elements are small.

	Parameters
	----------
	c: (d, d) np.array 
		Square input matrix for computing log-determinant.

	Returns
	-------
	d : float
		Log-determinant of the input matrix.
	"""
	u, s, v = svd(c)
	
	return np.sum(np.log(s))


def hqi(h, q):
	"""
	Computes :math:`\\bar{h}_q^{-1}(x)` where

	.. math::

		\\bar{h}_q(a) = -a \\log a -(1-a) \\log \\left(\\frac{1-a}{q-1}\\right), ~~~~ a \\geq \\frac{1}{q}.
	"""
	if h==0:
		return 1.0

	a = np.round(np.arange(1./q, 1.0, 0.001), 3)
	h_ = -a*np.log(a)-(1.-a)*np.log((1.-a)/(q-1.))
	idx = np.abs(h_-h).argmin()

	return a[idx]


def one_hot_encoding(x):
	"""
	.. _one-hot-encoding:
	Computes the `one hot encoding <https://en.wikipedia.org/wiki/One-hot>`_ representation of the input array.

	The representation used is the one where all distincts inputs are first converted to string, then sorted using Python's :code:`sorted` method. 

	The :math:`i`-th element in this sort has its :math:`i`-th bit from the right set to 1 and all others to 0.


	Parameters
	----------
	x: (n,) or (n, d) np.array
		Array of n inputs (typically strings) to encode and that take q distinct values.


	Returns
	-------
	 : (n, q) np.array
		Binary array representing the one-hot encoding representation of the inputs.
	"""
	x_ = np.array([str(_) for _ in x]).astype(str)
	n = len(x_)
	cats = set(x_)
	cats = list(sorted(cats))
	q = len(cats)
	res = np.zeros((n, q)).astype(int)

	for i in range(q):
		res[x_ == cats[i], q-1-i] = 1

	if len(res.shape) == 1:
		res = res[:, None]

	return res


def two_split_encoding(x):
	"""
	.. _two-split-encoding:
	Also known as binary encoding, the two-split encoding method turns categorical data taking :math:`q` distinct values into the ordinal data :math:`1, \\dots, q`, and then generates the binary representation of the ordinal data. 

	This encoding is more economical than one-hot encoding. Unlike the one-hot encoding methods which requires as many columns as the number of distinct inputs, two-split encoding only requires :math:`\\log_2 \\left\\lceil q \\right\\rceil` columns.

	Every bit in the encoding splits the set of all :math:`q` possible categories in 2 subsets of equal size (when q is a power of 2), and the value of the bit determines which subset the category of interest belongs to. Each bit generates a different partitioning of the set of all distinct categories.

	The ordinal value assigned to a categorical value is the order of its string representation among all :math:`q` distinct categorical values in the array.

	In other words, while a bit in one-hot encoding determines whether the category is equal to a specific values, a bit in the two-split encoding determines whether the category belong in one of two subsets of equal size.

	Parameters
	----------
	x: (n,) or (n, d) np.array
		Array of n inputs (typically strings) to encode and that take q distinct values.


	Returns
	-------
	 : (n, q) np.array
		Binary array representing the two-split encoding representation of the inputs.
	"""
	x_ = np.array([str(_) for _ in x]).astype(str)
	n = len(x_)
	cats = set(x_)
	cats = list(sorted(cats))
	q = len(cats)
	h = int(np.ceil(np.log2(q)))
	cats_map = {cats[i]: i for i in range(q)}
	res = np.array([[int(_) for _ in list(bin(cats_map[x_[i]])[2:].zfill(h))] for i in range(n)])

	if len(res.shape) == 1:
		res = res[:, None]

	return res


def get_y_data(y_c, y_d, categorical_encoding='two-split'):
	'''
	'''
	to_stack = []
	# Ouput
	if y_d is None:
		# Regression
		y_ = y_c[:, None] if len(y_c.shape) == 1 else y_c
		to_stack += [y_.copy()]
		output_indices = [0]
		n_outputs = y_.shape[1]

	if y_c is None:
		# Classification: threat the label like its encoded version
		y_ = one_hot_encoding(y_d) if categorical_encoding == 'one-hot' else two_split_encoding(y_d)
		to_stack += [y_.copy()]
		output_indices = [_ for _ in range(y_.shape[1])]
		n_outputs = y_.shape[1]

	z = np.hstack(to_stack).astype(float)

	return z, output_indices


def get_x_data(x_c, x_d, n_outputs, categorical_encoding='two-split', non_monotonic_extension=True, \
		space='dual'):
	'''
	'''
	to_stack = []
	n_inputs = 0
	n_features = 0

	x_c_ = None if x_c is None else x_c[:, None] if len(x_c.shape) == 1 else x_c
	x_d_ = None if x_d is None else x_d[:, None] if len(x_d.shape) == 1 else x_d
	batch_indices = []

	if x_c is not None:
		d = x_c_.shape[1]
		n_inputs += d
		to_stack += [x_c_]
		n_features += d
		if non_monotonic_extension and space=='dual':
			to_stack += [np.abs(x_c_-np.nanmean(x_c_, axis=0))]
			n_features += d
			batch_indices += [[i, i+d] for i in range(n_outputs, d+n_outputs)]

		else:
			batch_indices += [[i] for i in range(n_outputs, d+n_outputs)]

	z = np.hstack(to_stack)
	if x_d is not None:
		n_inputs += x_d_.shape[1]
		for j in range(x_d_.shape[1]):
			n_vars = z.shape[1] + n_outputs
			e = one_hot_encoding(x_d_[:, j]) if categorical_encoding == 'one-hot' else two_split_encoding(x_d_[:, j])
			dd = e.shape[1]
			# Threat the encoding of categorical variables as continuous
			z = e.copy() if z is None else np.hstack((z, e.copy()))
			if non_monotonic_extension and space=='dual':
				z = np.hstack((z, np.abs(e-np.nanmean(e, axis=0))))
				dd *= 2 
			n_features += dd
			batch_indices += [[_ for _ in range(n_vars, n_vars+dd)]]

	z = z.astype(float)
	# Parameter validation
	assert len(batch_indices) == n_inputs

	return z, batch_indices


def prepare_data_for_mutual_info_analysis(x_c, x_d, y_c, y_d, non_monotonic_extension=True, \
		categorical_encoding='two-split', space='dual'):
	"""
	"""
	from kxy.api.core import scalar_continuous_entropy
	assert y_c is None or np.can_cast(y_c, float), 'y_c should be an array of numeric values'
	assert y_c is None or len(y_c.shape) == 1 or y_c.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert y_d is None or len(y_d.shape) == 1 or y_d.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert x_c is None or np.can_cast(x_c, float), 'x_c should represent draws from a continuous random variable.'
	assert x_c is not None or x_d is not None, 'Both x_c and z_c cannot be None'
	assert y_c is not None or y_d is not None, 'Both y_c and y_d cannot be None'
	assert y_c is None or y_d is None, 'Both y_c and y_d cannot be not None'

	res = {}
	y_data, output_indices = get_y_data(y_c, y_d, categorical_encoding=categorical_encoding)
	n_outputs = y_data.shape[1]
	x_data, batch_indices = get_x_data(x_c, x_d, n_outputs, categorical_encoding=categorical_encoding, \
		non_monotonic_extension=non_monotonic_extension, space=space)
	all_data = np.hstack((y_data, x_data))

	# Constraints
	corr = pearson_corr(all_data) if space == 'primal' else spearman_corr(all_data)

	# Parameter validation
	assert not np.isnan(corr).any()

	# Putting everything together
	res['corr'] = corr
	res['batch_indices'] = batch_indices
	res['output_indices'] = output_indices
	res['x_data'] = x_data
	res['y_data'] = y_data

	return res


def empirical_copula_uniform(x):
	'''
	Evaluate the empirical copula-uniform dual representation of x as rank(x)/n.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from a d-dimensional distribution.

	'''
	mask = np.isnan(x).copy()
	valid_mask = np.logical_not(mask).astype(int)
	r = rankdata(x, axis=0)
	np.copyto(r, np.nan, where=mask)
	non_nan_ns = valid_mask.astype(float).sum(axis=0)
	u = r/non_nan_ns

	return u


def prepare_test_data_for_prediction(test_x_c, test_x_d, train_x_c, train_x_d, train_y_c, train_y_d, \
		non_monotonic_extension=True, categorical_encoding='two-split', space='dual'):
	"""
	"""
	train_res = prepare_data_for_mutual_info_analysis(\
		train_x_c, train_x_d, train_y_c, train_y_d, non_monotonic_extension=non_monotonic_extension, \
		categorical_encoding=categorical_encoding, space=space)

	train_x_data = train_res['x_data']
	train_y_data = train_res['y_data']
	n_outputs = train_y_data.shape[1]
	train_output_indices = train_res['output_indices']
	train_batch_indices = train_res['batch_indices']
	train_corr = train_res['corr']

	test_x_data, _ = get_x_data(test_x_c, test_x_d, n_outputs, categorical_encoding=categorical_encoding, \
		non_monotonic_extension=non_monotonic_extension, space=space)

	x = np.vstack((train_x_data, test_x_data))
	u_x = empirical_copula_uniform(x)
	test_u_x = u_x[train_x_data.shape[0]:, :]

	u_x_dict_list = [{n_outputs+i: row[i] for i in range(len(row)) \
		if not np.isnan(row[i]) and row[i] is not None} for row in test_u_x]

	res = {}
	res['corr'] = train_corr
	res['output_indices'] = train_output_indices
	res['batch_indices'] = train_batch_indices
	res['u_x_dict_list'] = u_x_dict_list
	res['problem_type'] = 'regression' if train_y_d is None else 'classification'
	
	if train_y_d is not None:
		train_y_data = train_y_data.astype(int)
		train_y_d_f = train_y_d.flatten()
	output_map = {str(train_y_data[i]): train_y_d_f[i] for i in range(len(train_y_d_f))} if not train_y_d is None else {}
	res['output_map'] = output_map
	res['train_y_data'] = train_y_data

	return res









#!/usr/bin/env python
# -*- coding: utf-8 -*-

from concurrent.futures import ThreadPoolExecutor
import numpy as np

import warnings
warnings.filterwarnings('ignore', module='statsmodels.tsa')
from statsmodels.tsa.api import VAR

from kxy.api.core import least_continuous_mutual_information, spearman_corr, \
	estimate_pearson_autocovariance, robust_log_det
from kxy.api import information_adjusted_correlation_from_spearman, \
	robust_pearson_corr_from_spearman


def information_adjusted_correlation(x, y=None, p=0):
	"""
	.. _information-adjusted-correlation:
	Calculates the information-adjusted correlation matrix between two arrays.

	.. note::

		Pearson's correlation coefficient quantifies the linear association between two random variables.
		Indeed, two random variables can be statistically dependent despite being decorrelated. Such dependence
		will typically materialize during tail events, the worst timing from a risk management perspective.


		The mutual information rate between two scalar time series provides an alternative that fully captures 
		linear and nonlinear, cross-sectional and temporal dependence:

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = h\\left( \\left\{x_t\\right\} \\right) +  h\\left( \\left\{y_t\\right\} \\right) - h\\left(\\left\{x_t, y_t\\right\}  \\right)

		where :math:`h\\left( \\left\{ x_t \\right\} \\right)` is the entropy rate of the process :math:`\\left\{ x_t \\right\}`. 

		Specifically, the mutual information rate is 0 if and only if the two processes are statistically independent, and in particular exhibit no
		cross-sectional or temporal dependence, linear or nonlinear.


		When :math:`\\left\{ x_t, y_t \\right\}` is Gaussian, stationary and memoryless, for instance when :math:`\\left(x_i, y_i \\right)`
		are assumed i.i.d Gaussian, the mutual entropy rate reads

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = -\\frac{1}{2} \log \\left(1- \\text{Corr}\\left(x_t, y_t\\right)^2 \\right).


		We generalize this formula and define the **information-adjusted correlation** as the quantity :math:`\\text{IACorr}\\left( \\left\{x_t\\right\}, \\left\{y_t\\right\} \\right)`
		so that the mutual information rate always reads

		.. math::
			I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) = -\\frac{1}{2} \log \\left(1- \\text{IACorr}\\left( \\left\{x_t\\right\}, \\left\{y_t\\right\} \\right)^2 \\right),

		whether or not the time series are jointly Gaussian and memoryless.

		.. math::
			\\text{IACorr}\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right) := \\text{sign}\\left( \\text{Corr}\\left(x_., y_.\\right) \\right)\\sqrt{1-e^{-2 I\\left(\\left\{x_t\\right\}, \\left\{y_t\\right\}\\right)}}

		where :math:`\\text{sign}(x)=1` if and only if :math:`x \geq 0` and :math:`-1` otherwise. Note that the information-adjusted correlation is 0
		if and only if the two time series are statistically independent, and in particular exhibit no cross-sectional or temporal dependencee.



	Parameters
	----------
	x : (n,) or (n, d) np.array
		n i.i.d. draws from a scalar or vector random variable.
	y : (n,) or (n, q) np.array
		n i.i.d. draws from a scalar or vector random variable jointly sampled with x.
	p : int
		The number of lags to use when generating Spearman rank auto-correlation to use 
		as empirical evidence in the maximum-entropy problem. The default value is 0, which 
		corresponds to assuming rows are i.i.d. This is also the only supported value for now.

		

	Returns
	-------
	c : np.array
		The information-adjusted correlation matrix between the two random variables. 

	Raises
	------
	AssertionError
		If p is different from 0. Higher values will be supported later.
	"""
	assert p==0, 'Only p=0 is supported for now'

	x_ = x[:, None] if len(x.shape) == 1 else x
	y_ = x_ if y is None else y[:, None] if len(y.shape) == 1 else y

	if p == 0:
		if (x_.shape == y_.shape) and np.allclose(x_, y_):
			corr = spearman_corr(x_)
			ia_corr = information_adjusted_correlation_from_spearman(corr)

		else:
			z = np.hstack([x_, y_])
			corr = spearman_corr(z)
			nx = x_.shape[1]
			ia_corr = information_adjusted_correlation_from_spearman(corr[:nx, nx:])
				
		return ia_corr




def robust_pearson_corr(x, y=None, p=0, p_ic='hqic'):
	"""
	.. _robust-pearson-corr:
	Computes a robust estimator of the Pearson correlation matrix between :math:`x` and :math:`y` (or :math:`x` if :math:`y` is None) as the Pearson correlation matrix that is equivalent to the sample Spearman correlation matrix, assuming :math:`(x, y)` is jointly Gaussian.


	Parameters
	----------
	x : (n,) or (n, d) np.array
		n i.i.d. draws from a scalar or vector random variable.
	y : (n,) or (n, q) np.array
		n i.i.d. draws from a scalar or vector random variable jointly sampled with x.
	p : int
		The number of lags to use when generating Spearman rank auto-correlation. 
		The default value is 0, which corresponds to assuming rows are i.i.d.
	p_ic : str
		The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. 
		Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). 
		Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.



	Returns
	-------
	c : np.array
		The robust Pearson correlation matrix between the two random variables. 
	"""
	x_ = x[:, None] if len(x.shape) == 1 else x
	y_ = x_ if y is None else y[:, None] if len(y.shape) == 1 else y
	is_auto_corr = (x_.shape == y_.shape) and np.allclose(x_, y_)
	z_ = x_ if is_auto_corr else np.hstack([x_, y_])

	if p == 0:
		if is_auto_corr:
			corr = spearman_corr(x_)
			rb_corr = robust_pearson_corr_from_spearman(corr)

		else:
			corr = spearman_corr(z_)
			nx = x_.shape[1]
			rb_corr = robust_pearson_corr_from_spearman(corr[:nx, nx:])

		return rb_corr



	else:
		if is_auto_corr:
			corr = spearman_corr(x_)

		else:
			nx = x_.shape[1]
			corr = spearman_corr(z_)[:nx, nx:]

		rb_corr = np.zeros_like(corr)
		if p is None:
			m = VAR(z_)
			try:
				p = m.fit(ic=p_ic).k_ar
			except:
				p = 1

		gamma_p = estimate_pearson_autocovariance(z_, p+1, robust=True)
		d = x_.shape[1]
		q = y_.shape[1]
		n = d if is_auto_corr else d+q

		def thread_mi_f(args):
			'''
			'''
			i, j = args
			if np.allclose(x_[:, i], y_[:, j]):
				return np.inf

			x_idx = list(i + np.arange(0, p+1).astype(int)*n)
			s = 0 if is_auto_corr else d
			y_idx = list(s + j + np.arange(0, p+1).astype(int)*n)
			xy_idx = sorted(x_idx + y_idx)

			gij_p = gamma_p[xy_idx].T[xy_idx].T
			huij = 0.5 * (robust_log_det(2. * np.pi * np.e * gij_p[:, :]) -\
						robust_log_det(2. * np.pi * np.e * gij_p[:-2, :-2]))

			gi_p = gamma_p[x_idx].T[x_idx].T
			hui = 0.5 * (robust_log_det(2. * np.pi * np.e * gi_p[:, :]) -\
						robust_log_det(2. * np.pi * np.e * gi_p[:-1, :-1]))

			gj_p = gamma_p[y_idx].T[y_idx].T
			huj = 0.5 * (robust_log_det(2. * np.pi * np.e * gj_p[:, :]) -\
						robust_log_det(2. * np.pi * np.e * gj_p[:-1, :-1]))

			return max(hui+huj-huij, 0.0)


		with ThreadPoolExecutor(max_workers=min(d*q, 100)) as pool:
			all_args = [(i, j) for j in range(q) for i in range(d)]
			results = [_ for _ in pool.map(thread_mi_f, all_args)]

		mis = np.zeros_like(corr)
		for k in range(len(all_args)):
			i, j = all_args[k]
			mis[i, j] = results[k]

		rb_corr = np.sqrt(1.-np.exp(-2.*mis))*np.sign(corr)

		return rb_corr




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from .base_accessor import BaseAccessor


@pd.api.extensions.register_dataframe_accessor("kxy_finance")
class FinanceAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for finance and asset management problems.

	This class defines the :code:`kxy_finance` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_finance.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""

	def beta(self, market_returns_column, asset_returns_columns=(), risk_free_column=None,\
			method='information-adjusted', p=0, p_ic='hqic'):
		"""
		Calculates the beta of a portfolio/asset (whose returns are provided in column_y) 
		with respect to the market (whose returns are provided in market_returns_column) using a variety
		of estimation methods including the standard OLS/Pearson methods and information theoretical 
		alternatives aiming at accounting for nonlinearities and memory in asset returns.


		Parameters
		----------
		asset_returns_columns : str or list of str
			The name(s) of the column(s) to use for portfolio/asset returns.
		market_returns_column : str
			The name of the column to use for market returns.
		method : str, optional
			One of 'information-adjusted', 'robust-pearson', 'spearman',  or 'pearson'. This is the method to use
			to estimate the correlation between portfolio/asset returns and market returns.
		p : int, optional
			The number of auto-correlation lags to use as empirical evidence in the maximum-entropy problem. 
			The default value is 0, which corresponds to assuming rows are i.i.d. Values other than 0 are only
			supported in the robust-pearson method. When p is None, it is inferred from the sample.
		p_ic : str
			The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`.
			Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion),
			'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). Same as the 'ic' parameter of 
			:code:`statsmodels.tsa.api.VAR`.


		Returns
		-------
		c : pandas.DataFrame
			The beta coefficient(s).


		.. seealso::

			:ref:`kxy.finance.factor_analysis.information_adjusted_beta <information-adjusted-beta>`
		"""
		asset_returns_columns = [_ for _ in self._obj.columns if _ != market_returns_column] if asset_returns_columns == () \
			else [asset_returns_columns] if type(asset_returns_columns) == str else list(asset_returns_columns)
		columns = [market_returns_column] + asset_returns_columns

		c = self.corr(method=method, columns=columns, p=p, p_ic=p_ic).values[0, 1:]
		betas = c * np.sqrt(np.nanvar(self._obj[asset_returns_columns].values, axis=0)/\
			np.nanvar(self._obj[market_returns_column].values))

		if type(asset_returns_columns) == str:
			res = pd.DataFrame({asset_returns_columns: betas}).T.rename(columns={0: 'beta'})
		else:
			res = pd.DataFrame({asset_returns_columns[i]: [betas[i]] \
				for i in range(len(asset_returns_columns))}).T.rename(columns={0: 'beta'})

		return res






from functools import wraps
import logging
import json

import numpy as np
import pandas as pd

from db import get_kxy_db_engine, get_session_factory, SpearmanCopulaDensityModel
from z import FullSpearmanZMC



def hits_api_first(method):
	'''
	The decorator seeks to get the reesult of the intended computation remotely first, 
	before attempting to run the computation locally.
	'''
	KXY_DB_ENGINE = get_kxy_db_engine()
	SESSION_FACTORY = get_session_factory()
	@wraps(method)
	def wrapper(*args, **kw):
		obj = args[0]

		if obj.__class__.__name__ == 'FullSpearmanMaxEntCopulaOptimizer' and method.__name__ == 'optimize':
			ordered_corr = FullSpearmanZMC.order_corr(obj.corr)
			db_session = SESSION_FACTORY()
			n_existing_copula = db_session.query(SpearmanCopulaDensityModel).filter_by(
				ordered_spearman_corr_matrix={i: list(ordered_corr[i]) for i in range(len(ordered_corr))}).count()

			if n_existing_copula and n_existing_copula > 1:
				existing_copula = db_session.query(SpearmanCopulaDensityModel).filter_by(
					ordered_spearman_corr_matrix={i: list(ordered_corr[i]) for i in range(len(ordered_corr))}).first()

				if existing_copula:
					logging.info('Found result from API for spearman correlation matrix %s ' % str(ordered_corr))
					logging.info('Setting cached result')
					z = FullSpearmanZMC.from_parameters_dict(existing_copula.theta, \
						existing_copula.d, n_moments=existing_copula.n_moments)

					obj.set_theta(z.theta)
					logging.info('Done setting cached result')
					return {'x': obj.theta}


		# Else run the method
		return method(*args, **kw)

	return wrapper





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from concurrent.futures import ThreadPoolExecutor
import json
import numpy as np
import requests

from kxy.api import APIClient, mutual_information_analysis


from .utils import prepare_data_for_mutual_info_analysis
from .entropy import least_structured_copula_entropy, least_structured_continuous_entropy, \
	least_structured_mixed_entropy, discrete_entropy


def least_total_correlation(x, space='dual'):
	""" 
	.. _least-total-correlation:
	Estimates the smallest total correlation of a continuous d-dimensional random vector,
	that is consistent with the average pairwise Spearman rank correlation estimated 
	from the input sample.

	.. math::
		TC(x) := \sum_{i=1}^d h(x_i)-h(x) = -h(u).

	This is the negative of :ref:`kxy.api.core.entropy.least_structured_copula_entropy <least-structured-copula-entropy>`.

	Parameters
	----------
	x : (n, d) np.array
		n i.i.d. draws from the data generating distribution.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.

	Returns
	-------
	tc : float
		The smallest total correlation consistent with observed empirical evidence, in nats.

	Raises
	------
	AssertionError
		If x is not a two dimensional array.
	"""
	assert len(x.shape) == 2
	return -least_structured_copula_entropy(x, space=space)


def least_mutual_information(x_c, x_d, y_c, y_d, space='dual', non_monotonic_extension=True, \
		categorical_encoding='two-split'):
	"""
	"""
	res = prepare_data_for_mutual_info_analysis(x_c, x_d, y_c, y_d, space=space, \
		non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)
	output_indices = res['output_indices']
	corr = res['corr']
	batch_indices = res['batch_indices']

	mi_ana = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)

	if mi_ana is None:
		return None

	return mi_ana['mutual_information']


def least_continuous_mutual_information(x_c, y, x_d=None, space='dual', non_monotonic_extension=True, \
		categorical_encoding='two-split'):
	"""
	.. _least-continuous-mutual-information:
	Estimates the mutual information between a :math:`d`-dimensional random vector :math:`x` of inputs
	and a continuous scalar random variable :math:`y` (or label), assuming the least amount of 
	structure in :math:`x` and :math:`(x, y)`, other than what is necessary to be consistent with some 
	observed properties.

	.. note::

		.. math::
			I(x, y) &= h(x) + h(y) - h(x, y) \\

			        &= h\\left(u_x\\right) - h\\left(u_{x,y}\\right)
		
		where :math:`u_x` (resp. :math:`u_{x,y}`) is the copula-uniform representation of :math:`x`
		(resp. :math:`(x, y)`).

		We use as model for :math:`u_{x, y}` the least structured copula that is consistent with 
		the Spearman rank correlation matrix of the joint vector :math:`(x, y)`. 

	Parameters
	----------
	x_c : (n,) or (n, d) np.array
		n i.i.d. draws from the continuous inputs data generating distribution.
	x_d : (n, q) np.array or None (default)
		n i.i.d. draws from the discrete inputs data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		When input parameters are invalid.
	"""
	return least_mutual_information(x_c, x_d, y, None, space=space, non_monotonic_extension=non_monotonic_extension, \
		categorical_encoding=categorical_encoding)



def least_continuous_conditional_mutual_information(x_c, y, z_c, x_d=None, z_d=None, space='dual',\
	non_monotonic_extension=True, categorical_encoding='two-split'):
	"""
	.. _least-continuous-conditional-mutual-information:
	Estimates the conditional mutual information between a :math:`d`-dimensional random vector :math:`x` of inputs
	and a continuous scalar random variable :math:`y` (or label), conditional on a third continuous random 
	variable :math:`z`, assuming the least amount of structure in :math:`(x, y, z)`, other than what is 
	necessary to be consistent with some observed properties.

	.. note::

		.. math::
			I(x; y|z) &= h(x|z) + h(y|z) - h(x, y|z) \\

			        &= h\\left(u_x | u_z \\right) - h\\left(u_{x,y} | u_z \\right)

			        &= I(y; x, z) - I(y; z)

		
		where :math:`u_x` (resp. :math:`u_{x,y}`, :math:`u_z`) is the copula-uniform representation of :math:`x`
		(resp. :math:`(x, y)`, :math:`z`).

		We use as model for :math:`u_{x, y, z}` the least structured copula that is consistent with 
		maximum entropy constraints in the chosen space. 

	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous inputs data generating distribution.
	x_d : (n, q) np.array or None (default)
		n i.i.d. draws from the discrete inputs data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (continuous) labels generating distribution, sampled
		jointly with x.
	z_c : (n, d) np.array
		n i.i.d. draws from the generating distribution of continuous conditions.
	z_d : (n, d) np.array or None (default), optional
		n i.i.d. draws from the generating distribution of categorical conditions.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		If y is not a one dimensional array or x, y and z are not all numeric.
	"""
	assert np.can_cast(y, float), 'y should be an array of numeric values'
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert x_c is None or np.can_cast(x_c, float), 'x_c should represent draws from a continuous random variable.'
	assert z_c is None or np.can_cast(z_c, float), 'x_c should represent draws from a continuous random variable.'
	assert x_c is not None or x_d is not None, 'Both x_c and z_c cannot be None'
	assert z_c is not None or z_d is not None, 'Both z_c and z_d cannot be None'

	x_c_ = None if x_c is None else x_c[:, None] if len(x_c.shape) == 1 else x_c
	z_c_ = None if z_c is None else z_c[:, None] if len(z_c.shape) == 1 else z_c
	x_d_ = None if x_d is None else x_d[:, None] if len(x_d.shape) == 1 else x_d
	z_d_ = None if z_d is None else z_d[:, None] if len(z_d.shape) == 1 else z_d

	# I(y; x_c, x_d | z_c, z_d) = I(y; x_c, x_d, z_c, z_d) - I(y; z_c, z_d)
	j_c = None if (x_c is None and z_c is None) else x_c_ if z_c is None else z_c_ if x_c is None else np.hstack((x_c_, z_c_)) # (x_c, z_c)
	j_d = None if (x_d is None and z_d is None) else x_d_ if z_d is None else z_d_ if x_d is None else np.hstack((x_d_, z_d_)) # (x_d, z_d)

	cmi = least_continuous_mutual_information(j_c, y, x_d=j_d, space=space, \
		non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)
	cmi -= least_continuous_mutual_information(z_c, y, x_d=z_d, space=space, \
		non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)

	return max(cmi, 0.0)




def least_mixed_mutual_information(x_c, y, x_d=None, space='dual', non_monotonic_extension=True, \
		categorical_encoding='two-split'):
	"""
	.. _least-mixed-mutual-information:
	Estimates the mutual inforrmation between some features (a d-dimensional random vector
	:math:`x_c` and possibly a discrete feature variable :math:`x_d` and a discrete output/label 
	random variable.	

	.. note::

		.. math::
			I({x_c, x_d}; y) &= h(y) + h(x_c, x_d) - h(x_c, x_d, y) \\

							 &= h(x_c, x_d) - h\\left(x_c, x_d \\vert y\\right) \\

							 &= h(x_c, x_d) - E\\left[h\\left(x_c, x_d \\vert y=.\\right)\\right] \\

							:&= h(x_c, x_d) - \sum_{j=1}^q \mathbb{P}(y=j)  h\\left(x_c, x_d \\vert y=j \\right)

		
		When there are discrete features:

		* :math:`h(x_c, x_d)` is estimated using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`,
		* :math:`\mathbb{P}(y=j)` is estimated using frequencies, 
		* :math:`h\\left(x_c, x_d | y=j\\right)` is estimated using :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

		When there are no discrete features, :math:`x_d` can simply be removed from the equations above, and :ref:`kxy.api.core.entropy.least_structured_continuous_entropy <least-structured-continuous-entropy>` is used for estimation instead of :ref:`kxy.api.core.entropy.least_structured_mixed_entropy <least-structured-mixed-entropy>`.

	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the continuous inputs data generating distribution.
	x_d : (n, q) np.array or None (default)
		n i.i.d. draws from the discrete inputs data generating distribution, jointly sampled with x_c, or None
		if there are no discrete features.
	y : (n,) np.array
		n i.i.d. draws from the (discrete) labels generating distribution, sampled jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.


	Returns
	-------
	i : float
		The mutual information between features and discrete labels, **in nats**.

	Raises
	------
	AssertionError
		If y is not a one-dimensional array, or x_c is not an array of numbers.
	"""
	return least_mutual_information(x_c, x_d, None, y, space=space, non_monotonic_extension=non_monotonic_extension, \
		categorical_encoding=categorical_encoding)




def least_mixed_conditional_mutual_information(x_c, y, z_c, x_d=None, z_d=None, space='dual', \
	non_monotonic_extension=True, categorical_encoding='two-split'):
	"""
	.. _least-mixed-conditional-mutual-information:
	Estimates the conditional mutual information between a dimensional random vector :math:`x` of inputs
	and a discrete scalar random variable :math:`y` (or label), conditional on a third random 
	variable :math:`z`, assuming the least amount of structure in :math:`(x, y, z)`, other than what is 
	necessary to be consistent with some observed properties.

	.. math::
		I(y; x_c, x_d | z_c, z_d) = I(y; x_c, x_d, z_c, z_d) - I(y; z_c, z_d)

		
	Parameters
	----------
	x_c : (n, d) np.array
		n i.i.d. draws from the generating distribution of continuous inputs.
	x_d : (n, d) np.array or None (default), optional
		n i.i.d. draws from the generating distribution of categorical inputs.
	z_c : (n, d) np.array
		n i.i.d. draws from the generating distribution of continuous conditions.
	z_d : (n, d) np.array or None (default), optional
		n i.i.d. draws from the generating distribution of categorical conditions.
	y : (n,) np.array
		n i.i.d. draws from the (categorical) labels generating distribution, sampled
		jointly with x.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	i : float
		The mutual information between (x_c, x_d) and y, conditional on (z_c, z_d), **in nats**.

	Raises
	------
	AssertionError
		If y is not a one dimensional array or x and z are not all numeric.
	"""
	assert x_c is not None or x_d is not None, 'Both x_c and z_c cannot be None'
	assert z_c is not None or z_d is not None, 'Both z_c and z_d cannot be None'
	assert x_c is None or np.can_cast(x_c, float), 'x_c should represent draws from a continuous random variable.'
	assert z_c is None or np.can_cast(z_c, float), 'x_c should represent draws from a continuous random variable.'

	x_c_ = None if x_c is None else x_c[:, None] if len(x_c.shape) == 1 else x_c
	z_c_ = None if z_c is None else z_c[:, None] if len(z_c.shape) == 1 else z_c
	x_d_ = None if x_d is None else x_d[:, None] if len(x_d.shape) == 1 else x_d
	z_d_ = None if z_d is None else z_d[:, None] if len(z_d.shape) == 1 else z_d

	# I(y; x_c, x_d | z_c, z_d) = I(y; x_c, x_d, z_c, z_d) - I(y; z_c, z_d)
	j_c = None if (x_c is None and z_c is None) else x_c_ if z_c is None else z_c_ if x_c is None else np.hstack((x_c_, z_c_)) # (x_c, z_c)
	j_d = None if (x_d is None and z_d is None) else x_d_ if z_d is None else z_d_ if x_d is None else np.hstack((x_d_, z_d_)) # (x_d, z_d)

	cmi = least_mixed_mutual_information(j_c, y.flatten(), x_d=j_d, space=space, \
		non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)
	cmi -= least_mixed_mutual_information(z_c, y.flatten(), x_d=z_d, space=space, \
		non_monotonic_extension=non_monotonic_extension, categorical_encoding=categorical_encoding)

	return max(cmi, 0.0)



def discrete_mutual_information(x, y):
	"""
	.. _discrete-mutual-information:
	Estimates the (Shannon) mutual information between two discrete random variables from
	i.i.d. samples, using the plug-in estimator of Shannon entropy.

	Parameters
	----------
	x : (n,) or (n, d) np.array or None (default)
		n i.i.d. draws from a discrete distribution.
	y : (n,) np.array
		n i.i.d. draws from another discrete distribution, sampled jointly with x.

	Returns
	-------
	i : float
		The mutual information between x and y, **in nats**.

	Raises
	------
	AssertionError
		If y or x is not a one-dimensional array, or if x and y have different shapes.
	"""
	assert len(y.shape) == 1 or y.shape[1] == 1, 'Only one-dimensional outputs are supported for now.'
	assert x.shape[0] == y.shape[0], 'Both arrays should have the same dimension.'

	x_ = x.flatten() if (len(x.shape) == 1) or (x.shape[1] == 1) else np.array(['*_*'.join([str(_) for _ in x[i]]) for i in len(x)])

	hx = discrete_entropy(x_)
	hy = discrete_entropy(y)
	hxy = discrete_entropy(np.array([str(x_[i]) + '*_*' + str(y[i]) for i in range(x.shape[0])]))

	return max(0., hx+hy-hxy)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from kxy.api import mutual_information_analysis
from kxy.api.core import prepare_data_for_mutual_info_analysis

from .pre_learning import regression_variable_selection_analysis, \
	regression_achievable_performance_analysis



def regression_model_improvability_analysis(x_c, y_p, y, x_d=None, space='dual', categorical_encoding='two-split'):
	"""
	.. _regression-model-improvability-analysis:
	Runs the model improvability analysis on a trained regression model.

	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	y_p : (n,) np.array
		Labels predicted by the model and corresponding to inputs x_c and x_d.
	y : (n,) np.array
		True labels.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Lost R^2'`: The amount by which the trained model's :math:`R^2` can still be improved without resorting to additional inputs, simply through better modeling.
		* :code:`'Lost Log-Likelihood Per Sample'`: The amount by which the trained model's true log-likelihood per sample can still be increased without resorting to additional inputs, simply through better modeling.
		* :code:`'Excess RMSE'`: The amount by which the trained model's RMSE can still be reduced without resorting to additional inputs, simply through better modeling.

	.. admonition:: Theoretical Foundation

		Section :ref:`3 - Model Improvability`.
	"""
	achievable_perf = regression_achievable_performance_analysis(x_c, y, x_d=x_d, space=space, categorical_encoding=categorical_encoding)
	achieved_perf   = regression_achievable_performance_analysis(y_p, y, x_d=None, space=space, categorical_encoding=categorical_encoding)
	
	improvable_perf = achievable_perf-achieved_perf
	improvable_perf['Achievable RMSE'] = -improvable_perf['Achievable RMSE']

	improvable_perf.rename(columns={col: col.replace('Achievable', 'Lost') for col in improvable_perf.columns}, inplace=True)
	improvable_perf.rename(columns={'Lost RMSE': 'Excess RMSE'}, inplace=True)
	improvable_perf = np.maximum(improvable_perf, 0.0)

	residual_perf = regression_achievable_performance_analysis(x_c, y-y_p, x_d=x_d, space=space, categorical_encoding=categorical_encoding)
	residual_perf.rename(columns={col: col.replace('Achievable', 'Residual') for col in residual_perf.columns}, inplace=True)
	residual_perf = np.maximum(residual_perf, 0.0)

	return pd.concat([improvable_perf, residual_perf], axis=1)



def regression_model_explanation_analysis(x_c, f_x, x_d=None, space='dual', categorical_encoding='two-split'):
	"""
	.. _regression-model-explanation-analysis:
	Runs the model explanation analysis on a trained regression model.

	Parameters
	----------
	x_c : (n,d) np.array
		Continuous inputs.
	f_x : (n,) np.array
		Labels predicted by the model and corresponding to inputs x_c and x_d.
	x_d : (n, d) np.array or None (default), optional
		Discrete inputs.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson covariance constraints, leading to the Gaussian copula.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank correlation constraints.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.

	Returns
	-------
	a : pandas.DataFrame

		Dataframe with columns:

		* :code:`'Variable'`: The variable index starting from 0 at the leftmost column of :code:`x_c` and ending at the rightmost column of :code:`x_d`.
		* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
		* :code:`'Univariate Explained R^2'`: The :math:`R^2` between predicted labels and this variable.
		* :code:`'Running Explained R^2'`: The :math:`R^2` between predicted labels and all variables selected so far, including this one.
		* :code:`'Marginal Explained R^2'`: The increase in :math:`R^2` between predicted labels and all variables selected so far that is due to adding this variable in the selection scheme.

	.. admonition:: Theoretical Foundation

		Section :ref:`a) Model Explanation`.
	"""
	res = regression_variable_selection_analysis(x_c, f_x, x_d=x_d, space=space, categorical_encoding=categorical_encoding)
	res = res[['Variable', 'Selection Order', 'Univariate Achievable R^2', 'Maximum Marginal R^2 Increase', \
		'Running Achievable R^2']]
	res.rename(columns={'Univariate Achievable R^2': 'Univariate Explained R^2', \
		'Maximum Marginal R^2 Increase': 'Marginal Explained R^2', \
		'Running Achievable R^2': 'Running Explained R^2'}, inplace=True)

	return res



def regression_bias(f_x, z, linear_scale=True, space='dual', categorical_encoding='two-split'):
	"""
	.. _regression-bias:
	Quantifies the bias in a regression model as the mutual information between a category variable and model predictions.


	Parameters
	----------
	f_x : (n,) np.array
		The model decisions.
	z : (n,) np.array
		The associated variable through which the bias could arise.
	categorical_encoding : str, 'one-hot' | 'two-split' (default)
		The encoding method to use to represent categorical variables. 
		See :ref:`kxy.api.core.utils.one_hot_encoding <one-hot-encoding>` and :ref:`kxy.api.core.utils.two_split_encoding <two-split-encoding>`.


	Returns
	-------
	b : float
		The mutual information :math:`m` or :math:`1-e^{-2m}` if :code:`linear_scale=True`.


	.. admonition:: Theoretical Foundation

		Section :ref:`b) Quantifying Bias in Models`.
	"""
	data = prepare_data_for_mutual_info_analysis(f_x, None, None, z, space=space, \
		non_monotonic_extension=True, categorical_encoding=categorical_encoding)
	output_indices = data['output_indices']
	corr = data['corr']
	batch_indices = data['batch_indices']

	mi_analysis = mutual_information_analysis(corr, output_indices, space=space, batch_indices=batch_indices)
	mi = mi_analysis['mutual_information']

	if linear_scale:
		mi = 1.-np.exp(-2.*mi)

	return mi





import hashlib
import threading
sampling_lock = threading.Lock()
from functools import lru_cache
import os
from time import time

import numpy as np
import numexpr as ne

from utils import cached_property

gU, gD, gM = None, None, None

class FullSpearmanZMC(object):
	'''
	'''
	def __init__(self, *args, **kwargs):
		self.theta = args[0]
		self.n_var = args[1] # Should be an int
		self.n_moments = kwargs.get('n_moments', 2)
		assert self.n_moments == 2, 'n_moments>2 not yet supported'
		assert len(self.theta) == self.n_moments * self.n_var + self.n_var * (self.n_var-1)//2
		self.random_state = np.random.RandomState(os.getpid())


	@cached_property
	def model_id(self):
		return ['d=%d' % self.n_var, 'n_moments=%d' % self.n_moments, 'theta=(%s)' % ','.join(['%.6f' % _ for _ in self.theta])]


	@cached_property
	def model_id_md5(self):
		return hashlib.md5(str(self.model_id).encode()).hexdigest()


	@property
	def theta_quad(self):
		'''
		Form the matrix A = a_ij such that a_ij is the coefficient of u_iu_j in the 
		'''
		res = np.zeros((self.n_var, self.n_var))
		shift = self.n_var
		for idx, i in enumerate(reversed(range(1, self.n_var+1))):
			res[idx, idx:idx+i] = self.theta[shift:shift+i]
			shift += i
		res = 0.5*(res + res.T)

		return res


	@property
	def theta_lin(self):
		'''
		First order coefficients (i.e. coefficient of u_i)
		'''
		return self.theta[:self.n_var]


	@cached_property
	def log_integrand_samples(self):
		''' 
		Generate samples from \theta^T * \phi(u)
		'''
		global gU, gM, gD

		if gU is None or gD != self.n_var or gM != self.n_moments:
			# Uniform samples
			U = self.random_state.rand(10000000, self.n_var)
			gU = U
			gM = self.n_moments
			gD = self.n_var

		else:
			U = gU

		# First order terms
		ret = np.dot(U, self.theta_lin)

		# Second order terms
		ret += np.sum(U*np.dot(self.theta_quad, U.T).T, axis=1)

		return ret, U


	@property
	def samples(self):
		''' 
		Generate samples from the variable exp(\theta^T * \phi(u)).
		'''
		samples_ = np.clip(self.log_integrand_samples[0], -500.0, 500.0)
		samples_ = ne.evaluate('exp(samples_)')  # np.exp()

		return samples_


	@property
	def gradient_samples(self):
		'''
		Generate samples from the variable \phi(u)[i] * exp(\theta^T * \phi(u)).
		'''
		U = self.log_integrand_samples[1]
		res = np.zeros((U.shape[0], self.theta.shape[0]))

		for i in range(self.n_var):
			res[:, i] = self.samples * U[:, i]

		shift = self.n_var
		for i, ni in enumerate(reversed(range(1, self.n_var+1))):
			for j in range(i, self.n_var):
				res[:, shift] = self.samples * U[:, i] * U[:, j]
				shift += 1

		return res


	@cached_property
	def samples_selector(self):
		return np.isfinite(self.samples)


	@cached_property
	def n_samples(self):
		return self.samples[self.samples_selector].shape[0]


	@cached_property
	def value(self):
		''' 
		Evaluate Z(theta) using Monte Carlo.
			Z(theta) = Avg(self.samples)
		'''
		return self.samples[self.samples_selector].mean()


	@cached_property
	def entropy(self):
		return np.log(self.value) - np.dot(self.gradient_log, self.theta)


	@cached_property
	def gradient_log(self):
		''' 
		Evaluate grad log Z(theta). 
		'''
		return self.gradient_samples[self.samples_selector].mean(axis=0)/self.value


	@cached_property
	def hessian_log(self):
		'''
		Evaluate the Hessian of log Z(theta). 
		'''
		DS = self.gradient_samples[self.samples_selector]
		D = DS.T / self.samples[self.samples_selector]
		res = (np.dot(D, DS)/self.n_samples)
		res /= self.value
		res -= np.outer(self.gradient_log, self.gradient_log)

		return res


	@cached_property
	def moment_constraints(self):
		'''
		Dictionary u_i^j: E[u_i^j]
		'''
		res = {'u_%d' % (i+1): self.gradient_log[i] for i in range(self.n_var)}
		shift = self.n_var
		for i, ni in enumerate(reversed(range(1, self.n_var+1))):
			res['u_%d^2' % (i+1)] = self.gradient_log[shift]
			shift += ni

		return res


	@cached_property
	def is_copula(self):
		'''
		Return False if and only if this is not the normalizing constant of a copula.
		'''
		if np.any(np.abs(self.gradient_log[:self.n_var]-0.5) > 0.001):
			return False

		shift = self.n_var
		for i, ni in enumerate(reversed(range(1, self.n_var+1))):
			if abs(self.gradient_log[shift]-1./3.) > 0.001:
				return False
			shift += ni

		return True


	@staticmethod
	def from_parameters_dict(theta, n_var, n_moments=2):
		'''
		Create an instance from saved parameters.
		'''
		theta = np.array(theta)
		return FullSpearmanZMC(theta, n_var, n_moments=n_moments)


	@cached_property
	def spearman_corr_matrix(self):
		'''
		Spearman correlation matrix. 
		rho_ij = 12E[u_ij]-3
		'''
		res = np.zeros((self.n_var, self.n_var))
		shift = self.n_var
		for i, ni in enumerate(reversed(range(1, self.n_var+1))):
			res[i, i:] = self.gradient_log[shift:shift+ni]
			res[i:, i] = self.gradient_log[shift:shift+ni]
			shift += ni

		res = 12.*res-3.0

		return res


	@staticmethod
	def spearman_corr_to_theta(corr, n_moments=2):
		'''
		'''
		n_var = corr.shape[0]
		theta = []
		theta += [0.5]*n_var
		shift = n_var
		for i, ni in enumerate(reversed(range(1, n_var+1))):
			theta += list((corr[i, i:]+3.)/12.)
			shift += ni

		if n_moments > 2:
			for m in range(3, n_moments+1):
				theta += [1./(m+1)]*n_var

		assert len(theta) == n_moments * n_var + n_var * (n_var-1)//2

		return np.array(theta)


	@staticmethod
	def order_corr(corr):
		n_var = corr.shape[0]
		indices = list(reversed(np.abs(corr).sum(axis=0).argsort()))
		ordered_corr = corr[indices].T[indices]
		rounded_ordered_corr = [['1.0' if i == j else '%.3f' % ordered_corr[i, j] \
					for j in range(n_var)] for i in range(n_var)]

		return rounded_ordered_corr		


	@cached_property
	def ordered_spearman_corr_matrix(self):
		'''
		Spearman correlation matrix, where variable are ordered by decreasing order of
		total absolute correllation.
		'''
		return FullSpearmanZMC.order_corr(self.spearman_corr_matrix)


	@cached_property
	def as_dict(self):
		return {\
			'created_at': time(), \
			'max_entropy_constraints': 'full-spearman-correlation-matrix', \
			'd': self.n_var, \
			'n_moments': self.n_moments, \
			'entropy': self.entropy, \
			'theta': list(self.theta), \
			'n_constraints': len(self.theta), \
			'expectations': list(self.gradient_log), \
			'spearman_corr_matrix': {i: list(self.spearman_corr_matrix[i, :]) for i in range(self.n_var)}, \
			'ordered_spearman_corr_matrix': {i: list(self.ordered_spearman_corr_matrix[i]) for i in range(self.n_var)}, \
			'moment_constraints': self.moment_constraints, \
			'model_id': str(self.model_id), \
			'model_id_md5': self.model_id_md5}


	@property
	def constraints_scales(self):
		return np.ones_like(self.theta)





import logging
import warnings

import numpy as np
from scipy.optimize import minimize

from db import get_session_factory, SpearmanCopulaDensityModel
from z import FullSpearmanZMC
from decorators import hits_api_first


class FullSpearmanMaxEntCopulaOptimizer(object):
	'''
	Distribution with minimum KL-divergence (maximum entropy) relative to the uniform distribution, 
	given the full Spearman rank correlation matrix, and with marginals approximately uniform 
	(same first n_moments moments as the standard uniform).
	'''
	def __init__(self, spearman_correlation_matrix=None, verbose=True, n_moments=2, record_copulas=False, optimize=True):
		assert spearman_correlation_matrix is not None, 'The spearman correlation matrix should be provided'
		assert spearman_correlation_matrix.shape[0] == spearman_correlation_matrix.shape[1], \
			'The spearman correlation matrix should be square'
		self.n_var = spearman_correlation_matrix.shape[0]
		self.n_moments = n_moments
		self.session_factory = get_session_factory()
		self.record_copulas = record_copulas
		self.corr = spearman_correlation_matrix
		self.constraints_values = FullSpearmanZMC.spearman_corr_to_theta(spearman_correlation_matrix, \
			n_moments=self.n_moments)
		self.optimize(verbose=verbose)


	@property
	def initial_theta(self):
		x0 = np.zeros_like(self.constraints_values)
		m = FullSpearmanMaxEntCopulaOptimizer(spearman_correlation_matrix=self.corr[:-1, :-1])
		x0[:self.n_var-1] = m.theta[:self.n_var-1]
		marker = self.n_var-1
		shift = self.n_var
		
		for idx, i in enumerate(reversed(range(1, self.n_var+1))):
			x0[shift:shift+i-1] = m.theta[marker:marker+i-1]
			marker += i-1
			shift += i
		
		return x0


	@hits_api_first
	def optimize(self, verbose=True):
		'''
		Learn the parameters of the minimum KL-divergence (maximum entropy)
	    relative to the uniform distribution, under constraints self.constraints.
		'''
		scale = 100.
		def f(theta):
			''' Squared norm of the difference between the linear constraints and their targets. '''
			z = FullSpearmanZMC(theta, self.n_var)
			e = z.gradient_log - self.constraints_values
			w = z.constraints_scales
			e = e * w

			if z.is_copula and self.record_copulas:
				d = z.as_dict.copy()
				c = SpearmanCopulaDensityModel(**d)
				try:
					db_session = self.session_factory()
					existing_copula = db_session.query(SpearmanCopulaDensityModel).filter_by(\
						model_id_md5=z.as_dict['model_id_md5']).first()
					if existing_copula:
						logging.info('Skipping previously recorded copula %s.' % (str(z.as_dict)))

					else:
						logging.info('Recording copula %s.' % (str(d)))
						db_session.add(c)
						db_session.commit()
						logging.info('Done recording copula %s.' % (str(d)))

				except:
					logging.exception("Failed to commit %s " % (str(d)))
					db_session.rollback()

				finally:
					if db_session:
						self.session_factory.remove()

			return scale*0.5*np.dot(e, e)


		def grad_f(theta):
			''' Gradient of the squared norm of the difference between the linear constraints and their targets. '''
			z = FullSpearmanZMC(theta, self.n_var)
			e = z.gradient_log - self.constraints_values
			w = z.constraints_scales
			e = e * (w**2)
			J = z.hessian_log

			return scale*np.dot(J, e)


		res = minimize(f, self.initial_theta, jac=grad_f)
		self.set_theta(res['x'])

		if verbose:
			logging.info(res)

		return res


	def set_theta(self, theta):
		'''
		Update distribution parameters
		'''
		self.theta = theta
		self.z = FullSpearmanZMC(self.theta, self.n_var)


if __name__ == '__main__':
	logging.basicConfig(format='%(asctime)s - %(process)d - %(levelname)s - %(message)s', level=logging.INFO)
	corr = np.array([[1.0, -0.378, 0.01], [-0.378, 1.0, 0.1], [0.01, 0.1, 1.0]])
	m = FullSpearmanMaxEntCopulaOptimizer(spearman_correlation_matrix=corr, record_copulas=True)






#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from .entropy_rate import gaussian_var_copula_entropy_rate


def least_continuous_mutual_information_rate(x, y, space='primal', robust=False, p=None, p_ic='hqic'):
	"""
	Estimate the maximum entropy mutual information rate between two scalar or vector valued time series.


	Parameters
	----------
	x : (T, d) np.array
		Sample path from the first time series.
	y : (T, q) np.array
		Sample path from the second time series.
	space : str
		One of :code:`'primal'` and :code:`'dual'` to choose where the maximum entropy optimization problem should be formulated.
	robust : bool
		If True and :code:`space='primal'`, then the autocovariance function is estimated by computing the Spearman rank autocorrelation
		function, inferring the equivalent autocorrelation function assuming Gaussianity, and then scaling back with the sample variances.
	p : int or None
		The number of autocorrelation lags to use for the maximum entropy problem. If set to :code:`None` (the default) and if :code:`space` is
		:code:`primal`, then it is inferred by fitting a VAR model on the joint time series using the Hannan-Quinn information criterion.
	p_ic : str
		The criterion used to learn the optimal value of :code:`p` (by fitting a VAR(p) model) when :code:`p=None`. 
		Should be one of 'hqic' (Hannan-Quinn Information Criterion), 'aic' (Akaike Information Criterion), 'bic' (Bayes Information Criterion) and 't-stat' (based on last lag). 
		Same as the 'ic' parameter of :code:`statsmodels.tsa.api.VAR`.
	space : str, 'primal' | 'dual'
		The space in which the maximum entropy problem is solved. 
		When :code:`space='primal'`, the maximum entropy problem is solved in the original observation space, under Pearson autocovariance constraints, leading to the Gaussian VAR.
		When :code:`space='dual'`, the maximum entropy problem is solved in the copula-uniform dual space, under Spearman rank autocorrelation constraints.


	.. warning::
		Maximum entropy optimization in the copula-uniform dual space is not yet supported for time series. 
		

	Returns
	-------
	i : float
		The mutual information rate between x and y, in nats.

	Raises
	------
	AssertionError
		If :code:`space` is not :code:`'primal'`.
	"""
	assert space == "primal", "Maximum-entropy optimization in the dual space is not yet implemented for time series"

	x_ = x if len(x.shape) > 1 else x[:, None]
	y_ = y if len(y.shape) > 1 else y[:, None]
	z_ = np.hstack((x_, y_))

	huxy, q = gaussian_var_copula_entropy_rate(z_, p=p, robust=robust, p_ic=p_ic)
	hux, _ = gaussian_var_copula_entropy_rate(x_, p=q, robust=robust, p_ic=p_ic)
	huy, _ = gaussian_var_copula_entropy_rate(y_, p=q, robust=robust, p_ic=p_ic)

	return max(hux+huy-huxy, 0.0)


